#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

"""
Constants for various key paths.
"""

import tempfile
from pathlib import Path

PACKAGE = Path(Path(__file__).resolve().parent)
"""The root directory of the package installation."""

SRC = Path(PACKAGE, r'..').resolve()
"""The root directory of repository's package sources."""

REPOSITORY = Path(SRC, r'..').resolve()
"""The root directory of the repository."""

TESTS = Path(REPOSITORY, r'tests')
"""The root directory of the repository's tests."""

IMG = Path(PACKAGE, r'img')
"""The img directory."""

CSS = Path(PACKAGE, r'css')
"""The css directory."""

JS = Path(PACKAGE, r'js')
"""The js directory."""

MCSS = Path(PACKAGE, r'mcss')
"""The root directory of the bundled m.css build."""

GENERATED = Path(PACKAGE, r'generated')
"""The root directory of all auto-generated package data."""

FONTS = Path(GENERATED, r'fonts')
"""Directory containing all the self-hosted google fonts."""

TEMP = Path(tempfile.gettempdir(), r'poxy')
"""A global temp directory shared by all instances of poxy."""

BUG_REPORT_DIR = (TEMP / r'bug_report').resolve()
"""Directory used for assembling bug reports."""

CASE_SENSITIVE = not (Path(str(PACKAGE).upper()).exists() and Path(str(PACKAGE).lower()).exists())
"""True if the file system is case-sensitive."""



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import copy
import os
from docutils.parsers import rst
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives, states
from docutils.nodes import fully_normalize_name, whitespace_normalize_name
from docutils.parsers.rst.roles import set_classes
from docutils import nodes

# If Pillow is not available, it's not an error unless one uses the image grid
# functionality (or :scale: option for Image)
try:
    import PIL.Image
    import PIL.ExifTags
except ImportError:
    PIL = None

default_settings = {
    'INPUT': None,
    'M_IMAGES_REQUIRE_ALT_TEXT': False
}

settings = None

class Image(Directive):
    """Image directive

    Copy of docutils.parsers.rst.directives.Image with:

    -   the align option removed (handled better by m.css)
    -   moved the implementation of scale here instead of it being handled in
        the HTML writer
    -   .m-image CSS class added
    -   adding a outer container for clickable image to make the clickable area
        cover only the image
    -   optionally dying when alt text is not present
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'alt': directives.unchanged_required,
                   'name': directives.unchanged,
                   'height': directives.length_or_unitless,
                   'width': directives.length_or_percentage_or_unitless,
                   'scale': directives.percentage,
                   'class': directives.class_option,
                   'target': directives.unchanged_required}

    # Overridden by Figure
    image_class = 'm-image'

    def run(self):
        messages = []
        reference = directives.uri(self.arguments[0])
        self.options['uri'] = reference
        reference_node = None
        if 'target' in self.options:
            block = states.escape2null(
                self.options['target']).splitlines()
            block = [line for line in block]
            target_type, data = self.state.parse_target(
                block, self.block_text, self.lineno)
            if target_type == 'refuri':
                reference_node = nodes.reference(refuri=data)
            elif target_type == 'refname':
                reference_node = nodes.reference(
                    refname=fully_normalize_name(data),
                    name=whitespace_normalize_name(data))
                reference_node.indirect_reference_name = data
                self.state.document.note_refname(reference_node)
            else:                           # malformed target
                messages.append(data)       # data is a system message
            del self.options['target']

        width = None
        height = None
        # If scaling requested, open the files and calculate the scaled size
        # Support both {filename} (3.7.1) and {static} (3.8) placeholders,
        # also prepend the absolute path in case we're not Pelican. In all
        # cases use only width and not both so the max-width can correctly
        # scale the image down on smaller screen sizes.
        # TODO: implement ratio-preserving scaling to avoid jumps on load using
        # the margin-bottom hack
        if 'scale' in self.options:
            file = os.path.join(os.getcwd(), settings['INPUT'])
            absuri = os.path.join(file, reference.format(filename=file, static=file))
            with PIL.Image.open(absuri) as im:
                width = "{}px".format(int(im.width*self.options['scale']/100.0))
        elif 'width' in self.options:
            width = self.options['width']
        elif 'height' in self.options:
            height = self.options['height'] # TODO: convert to width instead?

        # Remove the classes from the image element, will be added either to it
        # or to the wrapping element later
        set_classes(self.options)
        classes = self.options.get('classes', [])
        if 'classes' in self.options: del self.options['classes']
        if 'width' in self.options: del self.options['width']
        if 'height' in self.options: del self.options['height']
        image_node = nodes.image(self.block_text, width=width, height=height, **self.options)

        if not 'alt' in self.options and settings['M_IMAGES_REQUIRE_ALT_TEXT']:
            error = self.state_machine.reporter.error(
                    'Images and figures require the alt text. See the M_IMAGES_REQUIRE_ALT_TEXT option.',
                    image_node,
                    line=self.lineno)
            return [error]

        self.add_name(image_node)
        if reference_node:
            if self.image_class:
                container_node = nodes.container()
                container_node['classes'] += [self.image_class] + classes
                reference_node += image_node
                container_node += reference_node
                return messages + [container_node]
            else:
                reference_node += image_node
                return messages + [reference_node]
        else:
            if self.image_class: image_node['classes'] += [self.image_class] + classes
            return messages + [image_node]

class Figure(Image):
    """Figure directive

    Copy of docutils.parsers.rst.directives.Figure with:

    -   the align option removed (handled better by m.css)
    -   the redundant figwidth option removed (use width/scale/height from the
        Image directive instead)
    -   .m-figure CSS class added
    """

    option_spec = Image.option_spec.copy()
    option_spec['figclass'] = directives.class_option
    has_content = True

    image_class = None

    def run(self):
        figclasses = self.options.pop('figclass', None)
        (image_node,) = Image.run(self)
        if isinstance(image_node, nodes.system_message):
            return [image_node]
        figure_node = nodes.figure('', image_node)
        if figclasses:
            figure_node['classes'] += figclasses
        figure_node['classes'] += ['m-figure']

        if self.content:
            node = nodes.Element()          # anonymous container for parsing
            self.state.nested_parse(self.content, self.content_offset, node)
            first_node = node[0]
            if isinstance(first_node, nodes.paragraph):
                caption = nodes.caption(first_node.rawsource, '',
                                        *first_node.children)
                caption.source = first_node.source
                caption.line = first_node.line
                figure_node += caption
            elif not (isinstance(first_node, nodes.comment)
                      and len(first_node) == 0):
                error = self.state_machine.reporter.error(
                      'Figure caption must be a paragraph or empty comment, got %s' % type(first_node),
                      nodes.literal_block(self.block_text, self.block_text),
                      line=self.lineno)
                return [figure_node, error]
            if len(node) > 1:
                figure_node += nodes.legend('', *node[1:])
        return [figure_node]

# Adapter to accommodate breaking change in Pillow 7.2
# https://pillow.readthedocs.io/en/stable/releasenotes/7.2.0.html#moved-to-imagefiledirectory-v2-in-image-exif
def _to_numerator_denominator_tuple(ratio):
    if isinstance(ratio, tuple):
        return ratio
    else:
        return ratio.numerator, ratio.denominator

class ImageGrid(rst.Directive):
    has_content = True

    def run(self):
        grid_node = nodes.container()
        grid_node['classes'] += ['m-imagegrid', 'm-container-inflate']

        rows = [[]]
        total_widths = [0]
        for uri_caption in self.content:
            # New line, calculating width from 0 again
            if not uri_caption:
                rows.append([])
                total_widths.append(0)
                continue

            uri, _, caption = uri_caption.partition(' ')

            # Open the files and calculate the overall width
            # Support both {filename} (3.7.1) and {static} (3.8) placeholders,
            # also prepend the absolute path in case we're not Pelican
            file = os.path.join(os.getcwd(), settings['INPUT'])
            absuri = os.path.join(file, uri.format(filename=file, static=file))

            # If no caption provided, get EXIF info, if it's there
            with PIL.Image.open(absuri) as im:
                if not caption and hasattr(im, '_getexif') and im._getexif() is not None:
                    exif = {
                        PIL.ExifTags.TAGS[k]: v
                        for k, v in im._getexif().items()
                        if k in PIL.ExifTags.TAGS and len(str(v)) < 256
                    }

                    # Not all info might be present
                    caption = []
                    if 'FNumber' in exif:
                        numerator, denominator = _to_numerator_denominator_tuple(exif['FNumber'])
                        caption += ["F{}".format(float(numerator)/float(denominator))]
                    if 'ExposureTime' in exif:
                        numerator, denominator = _to_numerator_denominator_tuple(exif['ExposureTime'])
                        if int(numerator) > int(denominator):
                            caption += ["{} s".format(float(numerator)/float(denominator))]
                        else:
                            caption += ["{}/{} s".format(numerator, denominator)]
                    if 'ISOSpeedRatings' in exif:
                        caption += ["ISO {}".format(exif['ISOSpeedRatings'])]
                    caption = ', '.join(caption)

            # If the caption is `..`, it's meant to be explicitly disabled
            if caption == '..': caption = ''

            rel_width = float(im.width)/im.height
            total_widths[-1] += rel_width
            rows[-1].append((uri, rel_width, caption))

        for i, row in enumerate(rows):
            row_node = nodes.container()

            for uri, rel_width, caption in row:
                image_reference = rst.directives.uri(uri)
                image_node = nodes.image('', uri=image_reference)

                # <figurecaption> in case there's a caption
                if caption:
                    text_nodes, _ = self.state.inline_text(caption, self.lineno)
                    text_node = nodes.paragraph('', '', *text_nodes)
                    overlay_node = nodes.caption()
                    overlay_node.append(text_node)

                # Otherwise an empty <div>
                else: overlay_node = nodes.container()

                link_node = nodes.reference('', refuri=image_reference)
                link_node.append(image_node)
                link_node.append(overlay_node)
                wrapper_node = nodes.figure(width="{:.3f}%".format(rel_width*100.0/total_widths[i]))
                wrapper_node.append(link_node)
                row_node.append(wrapper_node)

            grid_node.append(row_node)

        return [grid_node]

def register_mcss(mcss_settings, **kwargs):
    global default_settings, settings
    settings = copy.deepcopy(default_settings)
    for key in settings.keys():
        if key in mcss_settings: settings[key] = mcss_settings[key]

    rst.directives.register_directive('image', Image)
    rst.directives.register_directive('figure', Figure)
    rst.directives.register_directive('image-grid', ImageGrid)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    settings = {
        'INPUT': pelicanobj.settings['PATH'],
    }
    for key in 'M_IMAGES_REQUIRE_ALT_TEXT':
        if key in pelicanobj.settings: settings[key] = pelicanobj.settings[key]

    register_mcss(mcss_settings=settings)

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and classes for working with SVG files.
"""

from typing import Sequence, Union

from lxml import etree

from . import xml_utils
from .utils import *


class SVG(object):
    pass

    def __init__(
        self,  #
        file_path: Union[Path, str],
        logger=None,
        root_id: str = None,
        id_namespace: str = None,
        root_classes: Union[str, Sequence[str]] = None,
    ):
        # read file
        svg = read_all_text_from_file(file_path, logger=logger)

        # add a namespace prefix to the #ids if requested
        # (so they're unique when injected into HTML etc)
        if root_id is not None:
            root_id = str(root_id).strip()
        if id_namespace is not None:
            id_namespace = str(id_namespace).strip()
        if root_id and id_namespace is None:
            id_namespace = root_id
        if id_namespace:
            svg = svg.replace(r'id="', rf'id="{id_namespace}-')
            svg = re.sub(r'''url\s*\(\s*(["']?)\s*#''', rf'url(\1#{id_namespace}-', svg, flags=re.I)
            svg = re.sub(r'''xlink:href\s*=\s*(["'])\s*#''', rf"xlink:href=\1#{id_namespace}-", svg, flags=re.I)

        # parse into XML
        self.__xml = xml_utils.read(svg, parser=xml_utils.create_parser(remove_blank_text=True))
        attrs = self.__xml.attrib

        # set/normalize various attributes
        if r'xmlns' not in attrs:
            attrs[r'xmlns'] = r'http://www.w3.org/2000/svg'
        # if r'xmlns:xlink' not in attrs:
        # 	attrs[r'xmlns:xlink'] = r'http://www.w3.org/1999/xlink')
        if r'version' not in attrs:
            attrs[r'version'] = r'1.1'
        if root_id:
            attrs[r'id'] = root_id

        # some editors use the root attribute to store a bunch of metadata
        # and tracking garbage so delete it
        if r'content' in attrs:
            del attrs[r'content']

        # set class attribute if specified
        if root_classes is not None:
            root_classes = list(coerce_collection(root_classes))
            if root_classes:
                attrs[r'class'] = r' '.join(list(coerce_collection(root_classes)))
            elif r'class' in attrs:
                del attrs[r'class']

    def __str__(self) -> str:
        return etree.tostring(self.__xml, encoding=r'unicode', xml_declaration=False, pretty_print=False)


__all__ = ['SVG']



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

from docutils.parsers.rst.states import Inliner
from docutils import nodes, utils
from docutils.parsers import rst
from docutils.parsers.rst.roles import set_classes

import xml.etree.ElementTree as ET
import os
import re

import logging

logger = logging.getLogger(__name__)

# Modified from abbr / gh / gl / ... to add support for queries and hashes
link_regexp = re.compile(r'(?P<title>.*) <(?P<link>[^?#]+)(?P<hash>[?#].+)?>')

def parse_link(text):
    link = utils.unescape(text)
    m = link_regexp.match(link)
    if m:
        title, link, hash = m.group('title', 'link', 'hash')
        if not hash: hash = '' # it's None otherwise
    else:
        title, hash = '', ''

    return title, link, hash

def init(tagfiles, input):
    global symbol_mapping, symbol_prefixes, tagfile_basenames

    # Pre-round to populate subclasses. Clear everything in case we init'd
    # before already.
    tagfile_basenames = []
    symbol_mapping = {}
    symbol_prefixes = ['']

    for f in tagfiles:
        tagfile, path = f[:2]
        prefixes = f[2] if len(f) > 2 else []
        css_classes = f[3] if len(f) > 3 else []

        tagfile_basenames += [(os.path.splitext(os.path.basename(tagfile))[0], path, css_classes)]
        symbol_prefixes += prefixes

        tree = ET.parse(os.path.join(input, tagfile))
        root = tree.getroot()
        for child in root:
            if child.tag == 'compound' and 'kind' in child.attrib:
                # Linking to pages
                if child.attrib['kind'] == 'page':
                    link = path + child.find('filename').text + '.html'
                    symbol_mapping[child.find('name').text] = (child.find('title').text, link, css_classes)

                # Linking to files
                if child.attrib['kind'] == 'file':
                    file_path = child.find('path')
                    link = path + child.find('filename').text
                    symbol_mapping[(file_path.text if file_path is not None else '') + child.find('name').text] = (None, link, css_classes)

                    for member in child.findall('member'):
                        if not 'kind' in member.attrib: continue

                        # Preprocessor defines and macros
                        if member.attrib['kind'] == 'define':
                            symbol_mapping[member.find('name').text + ('()' if member.find('arglist').text else '')] = (None, link + '#' + member.find('anchor').text, css_classes)

                # Linking to namespaces, structs and classes
                if child.attrib['kind'] in ['class', 'struct', 'namespace']:
                    name = child.find('name').text
                    # The cppreference tag file has <filename> empty
                    link = path + child.findtext('filename')
                    symbol_mapping[name] = (None, link, css_classes)
                    for member in child.findall('member'):
                        if not 'kind' in member.attrib: continue

                        # In case of the cppreference tag file, <compound>
                        # <filename> is empty, and <anchorfile> inside <member>
                        # is used instead. Doxygen fills both, so use
                        # <anchorfile> to cover both cases.
                        link = path + member.find('anchorfile').text

                        # Typedefs, constants, variables
                        if member.attrib['kind'] in ['typedef', 'enumvalue', 'variable']:
                            symbol_mapping[name + '::' + member.find('name').text] = (None, link + ('#' + member.findtext('anchor') if member.findtext('anchor') else ''), css_classes)

                        # Functions
                        if member.attrib['kind'] == 'function':
                            # <filename> can be empty (cppreference tag file)
                            symbol_mapping[name + '::' + member.find('name').text + "()"] = (None, link + ('#' + member.findtext('anchor') if member.findtext('anchor') else ''), css_classes)

                        # Enums with values
                        if member.attrib['kind'] == 'enumeration':
                            enumeration = name + '::' + member.find('name').text
                            symbol_mapping[enumeration] = (None, link + '#' + member.find('anchor').text, css_classes)

                            for value in member.findall('enumvalue'):
                                symbol_mapping[enumeration + '::' + value.text] = (None, link + '#' + value.attrib['anchor'], css_classes)

                # Sections. While rather strange, these reuse the `link`
                # variable defined by whatever came earlier --- because pages
                # have it with implicit `.html` but e.g. files with explicit
                # `.html`. TODO: fix more robustly
                for section in child.findall('docanchor'):
                    symbol_mapping[section.text] = (section.attrib.get('title', ''), link + '#' + section.text, css_classes)

def dox(name, rawtext, text, lineno, inliner: Inliner, options={}, content=[]):
    title, target, hash = parse_link(text)

    # Otherwise adding classes to the options behaves globally (uh?)
    _options = dict(options)
    set_classes(_options)
    # Avoid assert on adding to undefined member later
    if 'classes' not in _options: _options['classes'] = []

    # Try linking to the whole docs first
    for basename, url, css_classes in tagfile_basenames:
        if basename == target:
            if not title:
                # TODO: extract title from index page in the tagfile
                logger.warning("Link to main page `{}` requires a title".format(target))
                title = target

            _options['classes'] += css_classes
            node = nodes.reference(rawtext, title, refuri=url + hash, **_options)
            return [node], []

    for prefix in symbol_prefixes:
        if prefix + target in symbol_mapping:
            link_title, url, css_classes = symbol_mapping[prefix + target]
            if title:
                use_title = title
            elif link_title:
                use_title = link_title
            else:
                if link_title is not None:
                    logger.warning("Doxygen anchor `{}` has no title, using its ID as link title".format(target))

                use_title = target

            _options['classes'] += css_classes
            node = nodes.reference(rawtext, use_title, refuri=url + hash, **_options)
            return [node], []

    # TODO: print file and line
    #msg = inliner.reporter.warning(
        #'Doxygen symbol %s not found' % target, line=lineno)
    #prb = inliner.problematic(rawtext, rawtext, msg)
    if title:
        logger.warning("Doxygen symbol `{}` not found, rendering just link title".format(target))
        node = nodes.inline(rawtext, title, **_options)
    else:
        logger.warning("Doxygen symbol `{}` not found, rendering as monospace".format(target))
        node = nodes.literal(rawtext, target, **_options)
    return [node], []

def register_mcss(mcss_settings, **kwargs):
    rst.roles.register_local_role('dox', dox)

    init(input=mcss_settings['INPUT'],
         tagfiles=mcss_settings.get('M_DOX_TAGFILES', []))

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    settings = {
        # For backwards compatibility, the input directory is pelican's CWD
        'INPUT': os.getcwd(),
    }
    for key in ['M_DOX_TAGFILES']:
        if key in pelicanobj.settings: settings[key] = pelicanobj.settings[key]

    register_mcss(mcss_settings=settings)

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and classes for working with Doxygen.
"""

import itertools
import os
import re
import shutil
import subprocess
from typing import Tuple

from lxml import etree

from . import graph, xml_utils
from .utils import *

# =======================================================================================================================
# functions
# =======================================================================================================================


def mangle_name(name):
    '''
    A lightweight version of doxygen's escapeCharsInString()
    (see https://github.com/doxygen/doxygen/blob/master/src/util.cpp)
    '''
    assert name is not None
    name = name.replace('_', '__')
    name = name.replace(':', '_1')
    name = name.replace('/', '_2')
    name = name.replace('<', '_3')
    name = name.replace('>', '_4')
    name = name.replace('*', '_5')
    name = name.replace('&', '_6')
    name = name.replace('|', '_7')
    name = name.replace('.', '_8')
    name = name.replace('!', '_9')
    name = name.replace(',', '_00')
    name = name.replace(' ', '_01')
    name = name.replace('{', '_02')
    name = name.replace('}', '_03')
    name = name.replace('?', '_04')
    name = name.replace('^', '_05')
    name = name.replace('%', '_06')
    name = name.replace('(', '_07')
    name = name.replace(')', '_08')
    name = name.replace('+', '_09')
    name = name.replace('=', '_0a')
    name = name.replace('$', '_0b')
    name = name.replace('\\', '_0c')
    name = name.replace('@', '_0d')
    name = name.replace(']', '_0e')
    name = name.replace('[', '_0f')
    name = name.replace('#', '_0g')
    name = re.sub(r'[A-Z]', lambda m: '_' + m[0].lower(), name)
    return name


def format_for_doxyfile(val):
    if val is None:
        return ''
    elif isinstance(val, str):
        return '"' + val.replace('"', '\\"') + '"'
    elif isinstance(val, Path):
        return format_for_doxyfile(str(val))
    elif isinstance(val, bool):
        return r'YES' if val else r'NO'
    elif isinstance(val, (int, float)):
        return str(val)
    else:
        assert False


def path() -> Path:
    if not hasattr(path, "val"):

        def test_path(p):
            if not p:
                return None
            p = Path(p)
            if not p.exists() or not p.is_file() or not os.access(str(p), os.X_OK):
                return None
            return p.resolve()

        doxygen = None
        for name in (r'doxygen.exe', r'doxygen'):
            doxygen = test_path(shutil.which(name))
            if doxygen is not None:
                break

        if doxygen is None:
            for p in (
                'C:\\Program Files\\doxygen\\bin\\doxygen.exe',  #
                'C:\\Program Files (x86)\\doxygen\\bin\\doxygen.exe',
                r'/usr/local/bin/doxygen',
            ):
                try:
                    doxygen = test_path(p)
                    if doxygen is not None:
                        break
                except:
                    pass

        if doxygen is None:
            raise Error(rf'Could not find Doxygen on system path')

        path.val = doxygen
    return path.val


def raw_version_string() -> str:
    if not hasattr(raw_version_string, "val"):
        proc = subprocess.run([str(path()), r'--version'], capture_output=True, encoding=r'utf-8', check=True)
        val = proc.stdout.strip() if proc.stdout is not None else ''
        if not val and proc.stderr.strip():
            raise Error(rf'doxygen exited with error: {proc.stderr.strip()}')
        setattr(raw_version_string, 'val', val)
    return raw_version_string.val


def version() -> Tuple[int, int, int]:
    if not hasattr(version, "val"):
        val = raw_version_string()
        val = re.fullmatch(r'^\s*v?\s*([0-9]+)\s*[.]\s*([0-9]+)\s*[.]\s*([0-9]+)(?:[^0-9].*)?$', val, flags=re.I)
        assert val
        setattr(version, 'val', (int(val[1]), int(val[2]), int(val[3])))
    return version.val


def version_string() -> str:
    return rf'{version()[0]}.{version()[1]}.{version()[2]}'


# =======================================================================================================================
# Doxyfile
# =======================================================================================================================


class Doxyfile(object):
    def __init__(self, input_path=None, output_path=None, cwd=None, logger=None, flush_at_exit=True):
        self.__logger = logger
        self.__dirty = True
        self.__text = ''
        self.__autoflush = bool(flush_at_exit)
        self.__cwd = Path.cwd() if cwd is None else coerce_path(cwd).resolve()
        assert_existing_directory(self.__cwd)

        # the input + output
        self.__input_path = input_path
        if self.__input_path is not None:
            self.__input_path = coerce_path(self.__input_path)
        self.__output_path = output_path
        if self.__output_path is not None:
            self.__output_path = coerce_path(self.__output_path)

        # read in doxyfile
        if self.__input_path is not None:
            if not self.__input_path.is_file():
                raise Error(rf'{self.__input_path} was not a file')
            self.__text = read_all_text_from_file(self.__input_path, logger=self.__logger).strip()
            self.cleanup()  # expands includes

        # ...or generate one
        else:
            result = subprocess.run(
                [str(path()), r'-s', r'-g', r'-'], check=True, capture_output=True, cwd=self.__cwd, encoding='utf-8'
            )
            self.__text = result.stdout.strip()

        # simplify regex searches by ensuring there's always leading and trailing newlines
        self.__text = f'\n{self.__text}\n'

    def cleanup(self):
        if not self.__dirty:
            return
        if 1:
            log(self.__logger, rf'Cleaning doxyfile')
            # https://github.com/marzer/poxy/issues/33:
            self.__text = re.sub(r'_ENCODING\s*=\s*"([^"]*?)"', r'_ENCODING = \1', self.__text)
            result = subprocess.run(
                [str(path()), r'-s', r'-u', r'-'],
                check=True,
                capture_output=True,
                cwd=self.__cwd,
                encoding=r'utf-8',
                input=self.__text,
            )
            self.__text = result.stdout.strip()
        self.__dirty = False

    def flush(self):
        self.cleanup()
        if self.__output_path is not None:
            log(self.__logger, rf'Writing {self.__output_path}')
            with open(self.__output_path, 'w', encoding='utf-8', newline='\n') as f:
                f.write(self.__text)

    def hash(self):
        return sha1(self.__text)

    def contains(self, text):
        assert text is not None
        return self.__text.find(text) != -1

    def get_value(self, key, fallback=None):
        pattern = re.compile(rf'\n\s*{key}\s*=(.*?)\n', flags=re.S)
        text = self.__text
        m = pattern.search(text)
        while m:
            # doxygen allows values to appear multiple times and only accepts the last one, hence the loop
            text = text[m.end() :]
            n = pattern.search(text)
            if not n:
                break
            m = n
        if m:
            val = m[1].strip(' "')
            return val if val else fallback
        return fallback

    def get_boolean(self, key, fallback=False):
        val = self.get_value(key)
        if val is None:
            return fallback
        return val.upper() == 'YES'

    def append(self, *args, end='\n', sep=' '):
        self.__text = rf'{self.__text}{sep.join(args)}{end}'
        self.__dirty = True
        return self

    def add_value(self, key, value=None):
        if value is not None:
            if isinstance(value, (list, tuple, set)):
                if value:
                    for v in value:
                        if v is not None:
                            self.append(rf'{key:<23}+= {format_for_doxyfile(v)}')
            else:
                self.append(rf'{key:<23}+= {format_for_doxyfile(value)}')
            self.__dirty = True
        return self

    def set_value(self, key, value=None):
        if value is not None and isinstance(value, (list, tuple, set)):
            if not value:
                self.append(rf'{key:<23}=')
            else:
                first = True
                for v in value:
                    if first:
                        self.append(rf'{key:<23}=  {format_for_doxyfile(v)}')
                    else:
                        self.add_value(key, v)
                    first = False
        else:
            self.append(rf'{key:<23}=  {format_for_doxyfile(value)}')
        self.__dirty = True
        return self

    def get_text(self):
        return self.__text

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if traceback is None and self.__autoflush:
            self.flush()


class Bool(object):
    def __init__(self, value: bool):
        self.__value = bool(value)

    def __str__(self) -> str:
        return r'yes' if self.__value else r'no'


class Prot(object):
    def __init__(self, value: graph.AccessLevel):
        self.__value = value

    def __str__(self) -> str:
        return self.__value.name.lower()


class Virt(object):
    def __init__(self, value: bool):
        self.__value = bool(value)

    def __str__(self) -> str:
        return r'virtual' if self.__value else r'non-virtual'


# =======================================================================================================================
# XML <=> Graph
# =======================================================================================================================

COMPOUNDS = {r'dir', r'file', r'group', r'page', r'class', r'struct', r'union', r'concept', r'namespace'}
KINDS = {*COMPOUNDS, r'typedef', r'enum', r'enumvalue', r'variable', r'function', r'define'}
KINDS_TO_NODE_TYPES = {
    r'dir': graph.Directory,
    r'file': graph.File,
    r'group': graph.Group,
    r'class': graph.Class,
    r'struct': graph.Struct,
    r'union': graph.Union,
    r'concept': graph.Concept,
    r'namespace': graph.Namespace,
    r'typedef': graph.Typedef,
    r'enum': graph.Enum,
    r'enumvalue': graph.EnumValue,
    r'variable': graph.Variable,
    r'function': graph.Function,
    r'define': graph.Define,
    r'page': graph.Page,
    r'friend': graph.Friend,
}
NODE_TYPES_TO_KINDS = {t: k for k, t in KINDS_TO_NODE_TYPES.items()}
COMPOUND_NODE_TYPES = {KINDS_TO_NODE_TYPES[c] for c in COMPOUNDS}
VERSION = r'1.9.5'


def _ordered(*types) -> list:
    assert types is not None
    assert types
    types = [*types]
    types.sort(key=lambda t: t.__name__)
    types = tuple(types)
    return types


def _parse_xml_file(g: graph.Graph, path: Path, log_func=None):
    assert g is not None
    assert path is not None

    root = xml_utils.read(path, logger=log_func)

    if root.tag not in (r'doxygenindex', r'doxygen'):
        return

    def extract_subelement_text(elem, subelem_tag: str):
        assert elem is not None
        assert subelem_tag is not None
        subelem = elem.find(subelem_tag)
        if subelem is not None:
            return subelem.text
        return None

    def extract_qualified_name(elem):
        assert elem is not None
        for tag in (r'qualifiedname', r'compoundname'):
            n = elem.find(tag)
            if n is not None:
                n = n.text.strip()
            if n:
                return n
        return None

    def parse_structured_text(node: graph.Node, elem):
        nonlocal g
        # top-level text in the tag
        if elem.text:
            text = g.get_or_create_node(type=graph.Text, parent=node)
            text.text = elem.text
        # child <tags>
        for child_elem in elem:
            if child_elem.tag == r'para':
                para = g.get_or_create_node(type=graph.Paragraph, parent=node)
                parse_structured_text(para, child_elem)
            elif child_elem.tag == r'ref':
                ref = g.get_or_create_node(type=graph.Reference, parent=node)
                ref.text = child_elem.text
                ref.kind = child_elem.get(r'kindref')
                resource = g.get_or_create_node(id=child_elem.get(r'refid'), parent=ref)
                if child_elem.get(r'external'):
                    resource.type = graph.ExternalResource
                    resource.file = child_elem.get(r'external')
            else:
                markup = g.get_or_create_node(type=graph.ExpositionMarkup, parent=node)
                markup.tag = child_elem.tag
                attrs = [(k, v) for k, v in child_elem.attrib.items()]
                attrs.sort(key=lambda kvp: kvp[0])
                markup.extra_attributes = tuple(attrs)
                parse_structured_text(markup, child_elem)
            # text that came after the child <tag>
            if child_elem.tail:
                text = g.get_or_create_node(type=graph.Text, parent=node)
                text.text = child_elem.tail

    def parse_text_subnode(node: graph.Node, subnode_type, elem, subelem_tag: str):
        assert node is not None
        assert elem is not None
        assert subelem_tag is not None
        assert subnode_type is not None
        if subnode_type in node:
            return
        subelem = elem.find(subelem_tag)
        if subelem is None:
            return
        nonlocal g
        subnode = g.get_or_create_node(type=subnode_type, parent=node)
        parse_structured_text(subnode, subelem)

    def parse_brief(node: graph.Node, elem):
        parse_text_subnode(node, graph.BriefDescription, elem, r'briefdescription')

    def parse_detail(node: graph.Node, elem):
        parse_text_subnode(node, graph.DetailedDescription, elem, r'detaileddescription')

    def parse_initializer(node: graph.Node, elem):
        parse_text_subnode(node, graph.Initializer, elem, r'initializer')

    def parse_type(node: graph.Node, elem, resolve_auto_as=None):
        assert node is not None
        assert elem is not None
        if graph.Type in node:
            return
        type_elem = elem.find(r'type')
        if type_elem is None:
            return
        # extract constexpr, constinit, static, mutable etc out of the type if doxygen has leaked it
        while type_elem.text:
            text = rf' {type_elem.text} '
            match = re.search(
                r'\s(?:(?:const(?:expr|init|eval)|static|mutable|explicit|virtual|inline|friend)\s)+', text
            )
            if match is None:
                break
            type_elem.text = (text[: match.start()] + r' ' + text[match.end() :]).strip()
            if match[0].find(r'constexpr') != -1:
                node.constexpr = True
            if match[0].find(r'constinit') != -1:
                node.constinit = True
            if match[0].find(r'consteval') != -1:
                node.consteval = True
            if match[0].find(r'static') != -1:
                node.static = True
            if match[0].find(r'mutable') != -1:
                node.mutable = True
            if match[0].find(r'explicit') != -1:
                node.explicit = True
            if match[0].find(r'virtual') != -1:
                node.virtual = True
            if match[0].find(r'inline') != -1:
                node.inline = True
        if type_elem.text == r'auto' and resolve_auto_as is not None:
            type_elem.text = resolve_auto_as
        parse_text_subnode(node, graph.Type, elem, r'type')

    def parse_location(node: graph.Node, elem):
        location = elem.find(r'location')
        if location is None:
            return
        node.file = location.get(r'file')
        try:
            node.line = location.get(r'line')
        except:
            pass
        node.column = location.get(r'column')
        attrs = []
        for k, v in location.attrib.items():
            if k not in (r'file', r'line', r'column'):
                attrs.append((k, v))
        attrs.sort(key=lambda kvp: kvp[0])
        node.extra_attributes = tuple(attrs)

    # <compound>
    # (these are doxygen's version of 'forward declarations', typically found in index.xml)
    for compound in root.findall(r'compound'):
        if not compound.get(r'kind'):
            raise Error(rf"Malformed XML: <compound> tag missing attribute 'kind'")
        if compound.get(r'kind') == r'friend':
            raise Error(rf"Malformed XML: <compound> tag attribute 'kind' had unexpected value 'friend'")

        node = g.get_or_create_node(id=compound.get(r'refid'), type=KINDS_TO_NODE_TYPES[compound.get(r'kind')])

        if node.type is graph.File:  # files use their local name?? doxygen is so fucking weird
            node.local_name = tail(
                extract_subelement_text(compound, r'name').strip().replace('\\', r'/').rstrip(r'/'), r'/'
            )  #
        else:
            node.qualified_name = extract_subelement_text(compound, r'name')

        # <member>
        for member_elem in compound.findall(rf'member'):
            member_kind = member_elem.get(r'kind')
            if member_kind == r'enumvalue':
                continue
            member = g.get_or_create_node(
                id=member_elem.get(r'refid'), type=KINDS_TO_NODE_TYPES[member_kind], parent=node
            )
            name = extract_subelement_text(member_elem, r'name')
            if name:
                if member.type is graph.Define:
                    member.local_name = name
                    member.qualified_name = name
                elif node.type not in (graph.Directory, graph.File):
                    member.local_name = name
                    if node.qualified_name:
                        member.qualified_name = rf'{node.qualified_name}::{name}'

    # <compounddef>
    for compounddef in root.findall(r'compounddef'):
        if not compounddef.get(r'kind'):
            raise Error(rf"Malformed XML: <compounddef> tag missing attribute 'kind'")
        if compounddef.get(r'kind') == r'friend':
            raise Error(rf"Malformed XML: <compounddef> tag attribute 'kind' had unexpected value 'friend'")

        node = g.get_or_create_node(id=compounddef.get(r'id'), type=KINDS_TO_NODE_TYPES[compounddef.get(r'kind')])
        node.access_level = compounddef.get(r'prot')
        parse_brief(node, compounddef)
        parse_detail(node, compounddef)
        parse_initializer(node, compounddef)
        parse_location(node, compounddef)
        parse_type(node, compounddef)

        # qualified name
        qualified_name = extract_subelement_text(compounddef, r'qualifiedname')
        qualified_name = qualified_name.strip() if qualified_name is not None else r''
        if not qualified_name and node.type in (graph.Directory, graph.File):
            qualified_name = compounddef.find(r'location')
            qualified_name = qualified_name.get(r'file') if qualified_name is not None else r''
            qualified_name = qualified_name.rstrip(r'/')
        if not qualified_name:
            qualified_name = extract_qualified_name(compounddef)
        node.qualified_name = qualified_name

        # get all memberdefs in one flat list
        memberdefs = [compounddef]
        memberdefs += [s for s in compounddef.findall(r'sectiondef')]
        memberdefs = [s.findall(r'memberdef') for s in memberdefs]  # list of lists of memberdefs
        memberdefs = list(itertools.chain.from_iterable(memberdefs))  # list of memberdefs

        def get_memberdefs(kind: str):
            nonlocal memberdefs
            return [m for m in memberdefs if m.get(r'kind') == kind]

        # all <memberdefs>
        for elem in memberdefs:
            kind = elem.get(r'kind')
            member = g.get_or_create_node(id=elem.get(r'id'), type=KINDS_TO_NODE_TYPES[kind], parent=node)
            parse_brief(member, elem)
            parse_detail(member, elem)
            parse_initializer(member, elem)
            parse_location(member, elem)
            member.local_name = extract_subelement_text(elem, r'name')
            member.qualified_name = extract_qualified_name(elem)
            member.access_level = elem.get(r'prot')
            member.static = elem.get(r'static')
            member.const = elem.get(r'const')
            member.constexpr = elem.get(r'constexpr')
            member.consteval = elem.get(r'consteval')
            member.inline = elem.get(r'inline')
            member.explicit = elem.get(r'explicit')
            member.virtual = True if elem.get(r'virtual') == r'virtual' else None
            member.strong = elem.get(r'strong')
            member.definition = extract_subelement_text(elem, r'definition')

            # fix trailing return types in some situations (https://github.com/mosra/m.css/issues/94)
            trailing_return_type = None
            if kind == r'function':
                type_elem = elem.find(r'type')
                args_elem = elem.find(r'argsstring')
                if (type_elem is not None and type_elem.text) and (  #
                    args_elem is not None and args_elem.text and args_elem.text.find(r'decltype') == -1
                ):
                    match = re.search(r'^(.*?)\s*->\s*([a-zA-Z][a-zA-Z0-9_::*&<>\s]+?)\s*$', args_elem.text)
                    if match:
                        args_elem.text = str(match[1])
                        trailing_return_type = str(match[2]).strip()
                        trailing_return_type = re.sub(r'\s+', r' ', trailing_return_type)
                        trailing_return_type = re.sub(r'(::|[<>*&])\s+', r'\1', trailing_return_type)
                        trailing_return_type = re.sub(r'\s+(::|[<>*&])', r'\1', trailing_return_type)

            parse_type(member, elem, resolve_auto_as=trailing_return_type)

        # enums
        for elem in get_memberdefs(r'enum'):
            member = g.get_or_create_node(id=elem.get(r'id'), type=graph.Enum, parent=node)
            for value_elem in elem.findall(r'enumvalue'):
                value = g.get_or_create_node(id=value_elem.get(r'id'), type=graph.EnumValue, parent=member)
                value.access_level = value_elem.get(r'prot')
                value.local_name = extract_subelement_text(value_elem, r'name')
                parse_brief(value, value_elem)
                parse_detail(value, value_elem)
                parse_initializer(value, value_elem)
                parse_location(value, value_elem)

        # typedefs
        for elem in get_memberdefs(r'typedef'):
            member = g.get_or_create_node(id=elem.get(r'id'), type=graph.Typedef, parent=node)

        # vars
        for elem in get_memberdefs(r'variable'):
            member = g.get_or_create_node(id=elem.get(r'id'), type=graph.Variable, parent=node)

        # functions
        for elem in get_memberdefs(r'function'):
            member = g.get_or_create_node(id=elem.get(r'id'), type=graph.Function, parent=node)

        #

        # <inner(dir|file|class|namespace|page|group|concept)>
        for inner_suffix in (r'dir', r'file', r'class', r'namespace', r'page', r'group', r'concept'):
            for inner_elem in compounddef.findall(rf'inner{inner_suffix}'):
                inner = g.get_or_create_node(id=inner_elem.get(r'refid'), parent=node)
                if inner_suffix == r'class':
                    if inner.id.startswith(r'class'):
                        inner.type = graph.Class
                    elif inner.id.startswith(r'struct'):
                        inner.type = graph.Struct
                    elif inner.id.startswith(r'union'):
                        inner.type = graph.Union
                elif node.type in (graph.Class, graph.Struct, graph.Union) and inner_suffix == r'group':
                    inner.type = graph.MemberGroup
                else:
                    inner.type = KINDS_TO_NODE_TYPES[inner_suffix]
                if node.type is graph.Directory:
                    if inner.type is graph.Directory:
                        inner.qualified_name = inner_elem.text
                    else:
                        assert inner.type is graph.File
                        inner.qualified_name = rf'{node.qualified_name}/{inner_elem.text}'
                elif node.type in graph.CPP_TYPES and inner.type in graph.CPP_TYPES:
                    inner.qualified_name = inner_elem.text


def read_graph_from_xml(folder, log_func=None) -> graph.Graph:
    assert folder is not None
    folder = coerce_path(folder).resolve()
    g = graph.Graph()

    # parse files
    for path in get_all_files(folder, all=r"*.xml"):
        try:
            _parse_xml_file(g=g, path=path, log_func=log_func)
        except KeyError:
            raise
        except graph.GraphError as ex:
            raise graph.GraphError(rf'Parsing {path.name} failed: {ex}')
        except Exception as ex:
            raise Error(rf'Parsing {path.name} failed: {ex}')

    # deduce any missing qualified_names for C++ constructs
    again = True
    while again:
        again = False
        for namespace in g(graph.Namespace, graph.Class, graph.Struct, graph.Union, graph.Enum):
            if not namespace.qualified_name:
                continue
            for member in namespace(
                graph.Namespace,
                graph.Class,
                graph.Struct,
                graph.Union,
                graph.Variable,
                graph.Concept,
                graph.Enum,
                graph.EnumValue,
                graph.Function,
                graph.Typedef,
            ):
                if member.local_name and not member.qualified_name:
                    member.qualified_name = rf'{namespace.qualified_name}::{member.local_name}'
                    again = True

    # deduce any missing qualified_names for files and folders
    again = True
    while again:
        again = False
        for dir in g(graph.Directory):
            if not dir.qualified_name:
                continue
            for member in dir(graph.Directory, graph.File):
                if member.local_name and not member.qualified_name:
                    member.qualified_name = rf'{dir.qualified_name}/{member.local_name}'
                    again = True

    # add missing dir nodes + link file hierarchy
    for node in list(g(graph.Directory, graph.File)):
        sep = node.qualified_name.rstrip(r'/').rfind(r'/')
        if sep == -1:
            continue
        parent_path = node.qualified_name[:sep]
        if not parent_path:
            continue
        parent = None
        for dir in g(graph.Directory):
            if dir.qualified_name == parent_path:
                parent = dir
                break
        if parent is None:
            parent = g.get_or_create_node(type=graph.Directory)
        parent.qualified_name = parent_path
        parent.add(node)

    # resolve file links
    for node in g:
        if not node or node.type in (graph.Directory, graph.File, graph.ExternalResource) or not node.file:
            continue
        for file in g(graph.File):
            if file.qualified_name == node.file:
                file.add(node)

    g.validate()

    # replace doxygen's stupid nondeterministic IDs with something more robust
    id_remap = dict()

    def fix_ids(node: graph.Node) -> str:
        nonlocal id_remap
        assert node is not None
        assert node.type is not None

        # enum values are special - their ids always begin with the ID of their owning enum
        if node.type is graph.EnumValue:
            assert node.has_parent(graph.Enum)
            assert node.local_name
            parent = list(node(graph.Enum, parents=True))[0]
            id = re.sub(r'[/+!@#$%&*()+=.,{}<>;:?\[\]\^\-\\]+', r'_', node.local_name).rstrip(r'_')
            id = rf'{fix_ids(parent)}_{id}'
            id_remap[node.id] = id
            return id

        # if we don't have a qualified name then there's no meaningful transformation to do
        # we also don't transform functions because overloading makes them ambiguous (todo: handle functions)
        if not node.qualified_name or node.type is graph.Function:
            id_remap[node.id] = node.id
            return node.id

        id = re.sub(r'[/+!@#$%&*()+=.,{}<>;:?\[\]\^\-\\]+', r'_', node.qualified_name).rstrip(r'_')
        if len(id) > 128:
            id = sha1(id)
        id = rf'{node.type_name.lower()}_{id}'
        id_remap[node.id] = id
        return id

    # g = g.copy(id_transform=fix_ids)
    # g.validate()

    return g


def write_graph_to_xml(g: graph.Graph, folder: Path, log_func=None):
    assert folder is not None
    folder.mkdir(exist_ok=True, parents=True)

    def make_structured_text(elem, nodes):
        assert elem is not None
        assert nodes is not None

        # all the ones at the start that are just plain text get
        # concatenated and set as the main text of the root subelement
        if elem.text is None:
            elem.text = r''
        while nodes and nodes[0].type is graph.Text:
            elem.text = elem.text + nodes[0].text
            nodes.pop(0)

        # paragraphs/references/other exposition markup
        prev = None
        while nodes:
            if nodes[0].type is graph.Paragraph:
                para = xml_utils.make_child(elem, rf'para')
                para.text = nodes[0].text
                make_structured_text(
                    para, [n for n in nodes[0](graph.Paragraph, graph.Text, graph.Reference, graph.ExpositionMarkup)]
                )
                prev = para
            elif nodes[0].type is graph.ExpositionMarkup:
                assert nodes[0].tag
                markup = xml_utils.make_child(elem, nodes[0].tag)
                for k, v in nodes[0].extra_attributes:
                    markup.set(k, v)
                markup.text = nodes[0].text
                make_structured_text(
                    markup, [n for n in nodes[0](graph.Paragraph, graph.Text, graph.Reference, graph.ExpositionMarkup)]
                )
                prev = markup
            elif nodes[0].type is graph.Reference and nodes[0].is_parent:
                ref = xml_utils.make_child(elem, rf'ref', refid=nodes[0][0].id)
                ref.text = nodes[0].text
                if nodes[0].kind:
                    ref.set(r'kindref', nodes[0].kind)
                if nodes[0][0].type is graph.ExternalResource:
                    ref.set(r'external', nodes[0][0].file)
                prev = ref
            else:
                assert nodes[0].type in (graph.Text, graph.Reference)
                assert prev is not None
                if prev.tail is None:
                    prev.tail = r''
                prev.tail = prev.tail + nodes[0].text
            nodes.pop(0)

    def make_text_subnode(elem, subelem_tag: str, node: graph.Node, subnode_type):
        assert elem is not None
        assert subelem_tag is not None
        assert node is not None
        assert subnode_type is not None
        subelem = xml_utils.make_child(elem, subelem_tag)
        subelem.text = r''
        if subnode_type not in node:
            return
        text = [n for n in node(subnode_type)]  # list of BriefDescription
        text = [
            [i for i in n(graph.Paragraph, graph.Text, graph.Reference, graph.ExpositionMarkup)] for n in text
        ]  # list of lists
        text = list(itertools.chain.from_iterable(text))  # flattened list of Text/Paragraph/Reference
        if not text:
            return
        make_structured_text(subelem, text)

    def make_brief(elem, node: graph.Node):
        make_text_subnode(elem, r'briefdescription', node, graph.BriefDescription)

    def make_detail(elem, node: graph.Node):
        make_text_subnode(elem, r'detaileddescription', node, graph.DetailedDescription)

    def make_initializer(elem, node: graph.Node):
        make_text_subnode(elem, r'initializer', node, graph.Initializer)

    def make_type(elem, node: graph.Node):
        make_text_subnode(elem, r'type', node, graph.Type)

    def make_location(elem, node: graph.Node):
        subelem = None
        if node.type is graph.Directory:
            subelem = xml_utils.make_child(elem, rf'location', file=rf'{node.qualified_name}/')
        elif node.type is graph.File:
            subelem = xml_utils.make_child(elem, rf'location', file=rf'{node.qualified_name}')
        else:
            subelem = xml_utils.make_child(elem, rf'location', line=str(node.line), column=str(node.column))
            if node.file:
                subelem.set(r'file', node.file)
            else:
                files = [f for f in node(graph.File, parents=True)]
                if files and files[0].qualified_name:
                    subelem.set(r'file', files[0].qualified_name)
        for k, v in node.extra_attributes:
            subelem.set(k, v)

    # serialize the compound nodes
    for node in g(*COMPOUND_NODE_TYPES):
        if not node:
            continue
        assert node.qualified_name

        kind = NODE_TYPES_TO_KINDS[node.type]
        assert kind in COMPOUNDS

        path = Path(folder, rf'{node.id}.xml')
        root = etree.XML(
            rf'''<doxygen
						xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
						xsi:noNamespaceSchemaLocation="compound.xsd"
						version="{VERSION}"
						xml:lang="en-US">
					<!-- This file was created by Poxy - https://github.com/marzer/poxy -->
					<compounddef id="{node.id}" kind="{kind}" language="C++">
						<compoundname>{node.local_name if node.type is graph.File else node.qualified_name}</compoundname>
					</compounddef>
				</doxygen>''',
            xml_utils.DEFAULT_PARSER,
        )

        # create the root <compounddef>
        compounddef = root.find(r'compounddef')
        if node.type not in (graph.Namespace, graph.Directory, graph.File, graph.Concept):
            compounddef.set(r'prot', str(Prot(node.access_level)))

        # <includes>
        if node.type in (graph.Class, graph.Struct, graph.Union, graph.Concept):
            files = [f for f in g(graph.File) if (f and f is not node and node in f)]
            for f in files:
                assert f.local_name
                xml_utils.make_child(compounddef, rf'includes', local=r'no').text = f.local_name

        # create all the <sectiondefs>
        # (empty ones will be deleted at the end)
        sectiondefs = (
            # namespace/file sections:
            r'enum',
            r'typedef',
            r'var',
            r'func',
            # class/struct/union sections:
            r'public-type',
            r'protected-type',
            r'private-type',
            r'public-static-func',
            r'protected-static-func',
            r'private-static-func',
            r'public-func',
            r'protected-func',
            r'private-func',
            r'public-static-attrib',
            r'protected-static-attrib',
            r'private-static-attrib',
            r'public-attrib',
            r'protected-attrib',
            r'private-attrib',
            r'friend',
        )
        sectiondefs = {k: xml_utils.make_child(compounddef, r'sectiondef', kind=k) for k in sectiondefs}

        # enums
        enums = list(node(graph.Enum))
        enums.sort(key=lambda n: n.qualified_name)
        for member in enums:
            section = r'enum'
            if node.type in (graph.Class, graph.Struct, graph.Union):
                section = rf'{Prot(member.access_level)}-type'
            elem = xml_utils.make_child(
                sectiondefs[section],
                rf'memberdef',
                id=member.id,
                kind=r'enum',
                static=str(Bool(member.static)),
                strong=str(Bool(member.strong)),
                prot=str(Prot(member.access_level)),
            )
            make_type(elem, member)
            xml_utils.make_child(elem, r'name').text = member.local_name
            xml_utils.make_child(elem, r'qualifiedname').text = member.qualified_name
            for value in member(graph.EnumValue):
                value_elem = xml_utils.make_child(elem, rf'enumvalue', id=value.id, prot=str(Prot(value.access_level)))
                xml_utils.make_child(value_elem, r'name').text = value.local_name
                if graph.Initializer in value:
                    make_initializer(value_elem, value)
                make_brief(value_elem, value)
                make_detail(value_elem, value)
            make_brief(elem, member)
            make_detail(elem, member)
            xml_utils.make_child(elem, r'inbodydescription').text = r''  # todo
            make_location(elem, member)

        # typedefs
        typedefs = list(node(graph.Typedef))
        typedefs.sort(key=lambda n: n.qualified_name)
        for member in typedefs:
            section = r'typedef'
            if node.type in (graph.Class, graph.Struct, graph.Union):
                section = rf'{Prot(member.access_level)}-type'
            elem = xml_utils.make_child(
                sectiondefs[section],
                rf'memberdef',
                id=member.id,
                kind=r'typedef',
                static=str(Bool(member.static)),
                prot=str(Prot(member.access_level)),
            )
            make_type(elem, member)
            xml_utils.make_child(elem, r'definition').text = member.definition
            xml_utils.make_child(elem, r'argsstring')
            xml_utils.make_child(elem, r'name').text = member.local_name
            xml_utils.make_child(elem, r'qualifiedname').text = member.qualified_name
            make_brief(elem, member)
            make_detail(elem, member)
            xml_utils.make_child(elem, r'inbodydescription').text = r''  # todo
            make_location(elem, member)

        # variables
        variables = list(node(graph.Variable))
        if node.type in (graph.Class, graph.Struct, graph.Union):
            static_vars = [v for v in variables if v.static]
            static_vars.sort(key=lambda n: n.qualified_name)
            variables = static_vars + [v for v in variables if not v.static]
        else:
            variables.sort(key=lambda n: n.qualified_name)
        for member in variables:
            section = r'var'
            if node.type in (graph.Class, graph.Struct, graph.Union):
                section = rf'{Prot(member.access_level)}-{"static-" if member.static else ""}attrib'
            elem = xml_utils.make_child(
                sectiondefs[section],
                rf'memberdef',
                id=member.id,
                kind=r'variable',
                prot=str(Prot(member.access_level)),
                static=str(Bool(member.static)),
                constexpr=str(Bool(member.constexpr)),
                constinit=str(Bool(member.constinit)),
                mutable=str(Bool(member.strong)),
            )
            make_type(elem, member)
            xml_utils.make_child(elem, r'definition').text = member.definition
            xml_utils.make_child(elem, r'argsstring')
            xml_utils.make_child(elem, r'name').text = member.local_name
            xml_utils.make_child(elem, r'qualifiedname').text = member.qualified_name
            make_brief(elem, member)
            make_detail(elem, member)
            make_initializer(elem, member)
            xml_utils.make_child(elem, r'inbodydescription').text = r''  # todo
            make_location(elem, member)

        # functions
        functions = list(node(graph.Function))
        functions.sort(key=lambda n: n.qualified_name)
        for member in functions:
            section = r'func'
            if node.type in (graph.Class, graph.Struct, graph.Union):
                section = rf'{Prot(member.access_level)}-{"static-" if member.static else ""}func'
            elem = xml_utils.make_child(
                sectiondefs[section],
                rf'memberdef',
                id=member.id,
                kind=r'function',
                prot=str(Prot(member.access_level)),
                static=str(Bool(member.static)),
                const=str(Bool(member.const)),
                constexpr=str(Bool(member.constexpr)),
                consteval=str(Bool(member.consteval)),
                explicit=str(Bool(member.explicit)),
                inline=str(Bool(member.inline)),
                noexcept=str(Bool(member.noexcept)),
                virtual=str(Virt(member.virtual)),
            )
            make_type(elem, member)
            xml_utils.make_child(elem, r'name').text = member.local_name
            xml_utils.make_child(elem, r'qualifiedname').text = member.qualified_name
            make_brief(elem, member)
            make_detail(elem, member)
            xml_utils.make_child(elem, r'inbodydescription').text = r''  # todo
            make_location(elem, member)

        # <initializer> for concepts
        if node.type is graph.Concept:
            make_initializer(compounddef, node)

        # <briefdescription>, <detaileddescription>, <location>
        make_brief(compounddef, node)
        make_detail(compounddef, node)
        make_location(compounddef, node)

        # <listofallmembers>
        if node.type in (graph.Class, graph.Struct, graph.Union):
            listofallmembers = xml_utils.make_child(compounddef, rf'listofallmembers')
            listofallmembers.text = r''
            for member_type in _ordered(graph.Function, graph.Variable):
                for member in node(member_type):
                    member_elem = xml_utils.make_child(
                        listofallmembers,
                        rf'member',
                        refid=member.id,
                        prot=str(Prot(member.access_level)),
                        virtual=str(Virt(member.virtual)),
                    )
                    xml_utils.make_child(member_elem, r'scope').text = node.qualified_name
                    xml_utils.make_child(member_elem, r'name').text = member.local_name

        # add the inners
        for inner_type in _ordered(
            graph.Directory,  #
            graph.File,
            graph.Namespace,
            graph.Class,
            graph.Struct,
            graph.Union,
            graph.Concept,
            graph.Page,
            graph.Group,
            graph.MemberGroup,
        ):
            for inner_node in node(inner_type):
                if not inner_node:
                    continue
                assert inner_node.qualified_name

                kind = None
                if inner_node.type in (graph.Class, graph.Struct, graph.Union):
                    kind = r'class'
                elif inner_node.type is graph.MemberGroup:
                    kind = r'group'
                else:
                    kind = NODE_TYPES_TO_KINDS[inner_node.type]
                inner_elem = xml_utils.make_child(compounddef, rf'inner{kind}', refid=inner_node.id)
                if node.type not in (graph.Namespace, graph.Directory, graph.File, graph.Group, graph.Page):
                    inner_elem.set(r'prot', str(Prot(inner_node.access_level)))
                inner_elem.text = inner_node.qualified_name

        # prune empty <sectiondefs> etc
        for tag_name in (r'sectiondef',):
            for elem in list(compounddef.findall(tag_name)):
                if not len(elem):
                    elem.getparent().remove(elem)

        if log_func:
            log_func(rf'Writing {path}')
        xml_utils.write(root, path)

    # serialize index.xml
    if 1:
        path = Path(folder, rf'index.xml')
        root = etree.XML(
            rf'''<doxygenindex
						xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
						xsi:noNamespaceSchemaLocation="index.xsd"
						version="{VERSION}"
						xml:lang="en-US">
					<!-- This file was created by Poxy - https://github.com/marzer/poxy -->
				</xmlns>''',
            parser=xml_utils.DEFAULT_PARSER,
        )
        for node_type in _ordered(*COMPOUND_NODE_TYPES):
            for node in g(node_type):
                compound = xml_utils.make_child(
                    root, r'compound', refid=node.id, kind=NODE_TYPES_TO_KINDS[node.type]
                )  #
                xml_utils.make_child(compound, r'name').text = node.qualified_name
                if node.type is graph.Directory:
                    continue
                for child_type in _ordered(graph.Define, graph.Function, graph.Variable, graph.Enum):
                    children = list(node(child_type))
                    if child_type is graph.Variable and node.type in (graph.Class, graph.Struct, graph.Union):
                        static_vars = [c for c in children if c.static]
                        static_vars.sort(key=lambda n: n.qualified_name)
                        children = static_vars + [c for c in children if not c.static]
                    else:
                        children.sort(key=lambda n: n.qualified_name)
                    for child in children:
                        assert child.local_name
                        member = xml_utils.make_child(
                            compound, r'member', refid=child.id, kind=NODE_TYPES_TO_KINDS[child.type]  #
                        )
                        xml_utils.make_child(member, r'name').text = child.local_name
                        if child_type is graph.Enum:
                            for enumvalue in child(graph.EnumValue):
                                assert enumvalue.local_name
                                elem = xml_utils.make_child(
                                    compound, r'member', refid=enumvalue.id, kind=NODE_TYPES_TO_KINDS[enumvalue.type]  #
                                )
                                xml_utils.make_child(elem, r'name').text = enumvalue.local_name

        if log_func:
            log_func(rf'Writing {path}')
        xml_utils.write(root, path)



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re

from docutils import nodes, utils
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.parsers.rst.roles import set_classes

import matplotlib as mpl
mpl.use('Agg') # otherwise it will attempt to use X11
import matplotlib.pyplot as plt
import numpy as np
import io

mpl.rcParams['font.size'] = '11'
mpl.rcParams['axes.titlesize'] = '13'

# Plot background. Replaced with .m-plot .m-background later, equivalent to
# --default-filled-background-color
mpl.rcParams['axes.facecolor'] = '#cafe01'

# All of these should match --color, replaced with .m-plot .m-text
mpl.rcParams['text.color'] = '#cafe02'
mpl.rcParams['axes.labelcolor'] = '#cafe02'
mpl.rcParams['xtick.color'] = '#cafe02'
mpl.rcParams['ytick.color'] = '#cafe02'

# no need to have a border around the plot
mpl.rcParams['axes.spines.left'] = False
mpl.rcParams['axes.spines.right'] = False
mpl.rcParams['axes.spines.top'] = False
mpl.rcParams['axes.spines.bottom'] = False

mpl.rcParams['svg.fonttype'] = 'none' # otherwise it renders text to paths
mpl.rcParams['figure.autolayout'] = True # so it relayouts everything to fit

# Gets increased for every graph on a page to (hopefully) ensure unique SVG
# IDs. The parameter is a number but "Support for setting an rcParam that
# expects a str value to a non-str value is deprecated since 3.5 and support
# will be removed two minor releases later", so we pass a string.
mpl.rcParams['svg.hashsalt'] = '0'

# Color codes for bars. Keep in sync with latex2svgextra.
style_mapping = {
    'default': '#cafe03',
    'primary': '#cafe04',
    'success': '#cafe05',
    'warning': '#cafe06',
    'danger': '#cafe07',
    'info': '#cafe08',
    'dim': '#cafe09'
}

# Patch to remove preamble and hardcoded sizes. Matplotlib 2.2 has a http URL
# while matplotlib 3 has a https URL, check for both. Matplotlib 3.3 has a new
# <metadata> field (which we're not interested in) and slightly different
# formatting of the global style after (which we unify to the compact version).
# Matplotlib 3.4 drops the "Created with" comment, as that's in the <metadata>
# already anyway. Matplotlib 3.5 changes order of the <svg> attributes (which
# we ignore anyway, so those aren't matched anymore), order of stroke CSS
# properties and whitespace/semicolons (which we preserve for consistency with
# the rest of the output).
_patch_src = re.compile(r"""<\?xml version="1\.0" encoding="utf-8" standalone="no"\?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1\.1//EN"
  "http://www\.w3\.org/Graphics/SVG/1\.1/DTD/svg11\.dtd">(
<!-- Created with matplotlib \(https?://matplotlib.org/\) -->)?
<svg [^>]+(?P<viewBox>viewBox="0 0 \d+ \d+(\.\d+)?")[^>]+>(
 <metadata>.+</metadata>)?
 <defs>
  <style type="text/css">
?\*{(?P<style>stroke-linecap:butt;stroke-linejoin:round;|stroke-linejoin: round; stroke-linecap: butt)}(
  )?</style>
 </defs>
""", re.DOTALL)
_patch_dst = r"""<svg \g<viewBox>>
 <defs>
  <style type="text/css">*{\g<style>}</style>
 </defs>
"""

# Remove needless newlines and trailing space in path data
_path_patch_src = re.compile('(?P<prev>[\\dz]) ?\n(?P<next>[LMz])', re.MULTILINE)
_path_patch_dst = '\\g<prev> \\g<next>'
_path_patch2_src = re.compile(' ?\n"')
_path_patch2_dst = '"'

# Mapping from color codes to CSS classes. Matplotlib 3.5 added quite a lot of
# spacing to the style attributes compared to previous versions (and dropped
# trailing semicolons), so there's a lot of ` ?` and ` ;` to cover both
# variants.
_class_mapping = [
    # Graph background
    (re.compile('style="fill: ?#cafe01;?"'), 'class="m-background"'),

    # Tick <path> definition in <defs>
    (re.compile('style="stroke: ?#cafe02; ?stroke-width: ?0.8;?"'), 'class="m-line"'),
    # <use>, everything is defined in <defs>, no need to repeat. Matplotlib 3.5
    # has style at the end of the tag, so preserving whatever is in between.
    (re.compile('<use([^>]*) style="fill: ?#cafe02; ?stroke: ?#cafe02; ?stroke-width: ?0.8;?"'), '<use\\1'),

    # Text styles have `font-stretch:normal;` added in matplotlib 3.3;
    # 3.5 changes all font properties to just a single `font:` declaration.

    # Label text on left
    (re.compile('style="fill: ?#cafe02;(font-family:[^;]+;font-size:11px;(font-stretch:normal;)?font-style:normal;font-weight:normal;| font: 11px \'[^\']+\')"'), 'class="m-label"'),
    # Label text on bottom (has extra style params)
    (re.compile('style="fill: ?#cafe02;(font-family:[^;]+;font-size:11px;(font-stretch:normal;)?font-style:normal;font-weight:normal;| font: 11px \'[^\']+\'; )'), 'class="m-label" style="'),
    # Secondary label text
    (re.compile('style="fill: ?#cafe0b;(font-family:[^;]+;font-size:11px;(font-stretch:normal;)?font-style:normal;font-weight:normal;| font: 11px \'[^\']+\')"'), 'class="m-label m-dim"'),
    # Title text (has extra style params)
    (re.compile('style="fill: ?#cafe02;(font-family:[^;]+;font-size:13px;(font-stretch:normal;)?font-style:normal;font-weight:normal;| font: 13px \'[^\']+\'; )'), 'class="m-title" style="'),

    # Bar colors. Keep in sync with latex2svgextra.
    (re.compile('style="fill: ?#cafe03;?"'), 'class="m-bar m-default"'),
    (re.compile('style="fill: ?#cafe04;?"'), 'class="m-bar m-primary"'),
    (re.compile('style="fill: ?#cafe05;?"'), 'class="m-bar m-success"'),
    (re.compile('style="fill: ?#cafe06;?"'), 'class="m-bar m-warning"'),
    (re.compile('style="fill: ?#cafe07;?"'), 'class="m-bar m-danger"'),
    (re.compile('style="fill: ?#cafe08;?"'), 'class="m-bar m-info"'),
    (re.compile('style="fill: ?#cafe09;?"'), 'class="m-bar m-dim"'),

    # Error bar line
    (re.compile('style="fill: ?none; ?stroke: ?#cafe0a; ?stroke-width: ?1.5;?"'), 'class="m-error"'),
    # Error bar <path> definition in <defs>
    (re.compile('style="stroke: ?#cafe0a;?"'), 'class="m-error"'),
    # <use>, everything is defined in <defs>, no need to repeat
    (re.compile('<use([^>]*) style="fill: ?#cafe0a; ?stroke: ?#cafe0a;?"'), '<use\\1'),
]

# Titles for bars
_bar_titles_src = '<g id="plot{}-value{}-{}">'
_bar_titles_dst = '<g id="plot{}-value{}-{}"><title>{} {}</title>'
_bar_titles_dst_error = '<g id="plot{}-value{}-{}"><title>{} ± {} {}</title>'

class Plot(rst.Directive):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'class': directives.class_option,
                   'name': directives.unchanged,
                   'type': directives.unchanged_required,
                   'labels': directives.unchanged_required,
                   'labels-extra': directives.unchanged,
                   'units': directives.unchanged_required,
                   'values': directives.unchanged_required,
                   'errors': directives.unchanged,
                   'colors': directives.unchanged,
                   'plot-width': directives.unchanged,
                   'bar-height': directives.unchanged,
                   # Legacy options with ugly underscores instead of dashes
                   'labels_extra': directives.unchanged,
                   'bar_height': directives.unchanged}
    has_content = False

    def run(self):
        set_classes(self.options)

        # Type
        assert self.options['type'] == 'barh'

        # Graph title and axis labels. Value labels are one per line.
        title = self.arguments[0]
        units = self.options['units']
        labels = self.options['labels'].split('\n')

        # Legacy options, convert underscores to dashes
        if 'labels_extra' in self.options:
            self.options['labels-extra'] = self.options['labels_extra']
            del self.options['labels_extra']
        if 'bar_height' in self.options:
            self.options['bar-height'] = self.options['bar_height']
            del self.options['bar_height']

        # Optional extra labels
        if 'labels-extra' in self.options:
            labels_extra = self.options['labels-extra'].split('\n')
            assert len(labels_extra) == len(labels)
        else:
            labels_extra = None

        # Values. Should be one for each label, if there are multiple lines
        # then the values get stacked.
        value_sets = []
        for row in self.options['values'].split('\n'):
            values = [float(v) for v in row.split()]
            assert len(values) == len(labels)
            value_sets += [values]

        # Optional errors
        if 'errors' in self.options:
            error_sets = []
            for row in self.options['errors'].split('\n'):
                errors = [float(e) for e in row.split()]
                assert len(errors) == len(values)
                error_sets += [errors]
            assert len(error_sets) == len(value_sets)
        else:
            error_sets = [None]*len(value_sets)

        # Colors. Should be either one for all or one for every value
        if 'colors' in self.options:
            color_sets = []
            for row in self.options['colors'].split('\n'):
                colors = [style_mapping[c] for c in row.split()]
                if len(colors) == 1: colors = colors[0]
                else: assert len(colors) == len(labels)
                color_sets += [colors]
            assert len(color_sets) == len(value_sets)
        else:
            color_sets = [style_mapping['default']]*len(value_sets)

        # Bar height
        bar_height = float(self.options.get('bar-height', '0.4'))

        # Increase hashsalt for every plot to ensure (hopefully) unique SVG
        # IDs. The parameter is a number but "Support for setting an rcParam
        # that expects a str value to a non-str value is deprecated since 3.5
        # and support will be removed two minor releases later", so we convert
        # back to a string.
        mpl.rcParams['svg.hashsalt'] = str(int(mpl.rcParams['svg.hashsalt']) + 1)

        # Setup the graph
        fig, ax = plt.subplots()
        # TODO: let matplotlib calculate the height somehow
        fig.set_size_inches(float(self.options.get('plot-width', 8)), 0.78 + len(labels)*bar_height)
        yticks = np.arange(len(labels))
        left = np.array([0.0]*len(labels))
        for i in range(len(value_sets)):
            plot = ax.barh(yticks, value_sets[i], xerr=error_sets[i],
                           align='center', color=color_sets[i], ecolor='#cafe0a', capsize=5*bar_height/0.4, left=left)
            left += np.array(value_sets[i])
            for j, v in enumerate(plot):
                v.set_gid('plot{}-value{}-{}'.format(mpl.rcParams['svg.hashsalt'], i, j))
        ax.set_yticks(yticks)
        ax.invert_yaxis() # top-to-bottom
        ax.set_xlabel(units)
        ax.set_title(title)

        # Value labels. If extra label is specified, create two multiline texts
        # with first having the second line empty and second having the first
        # line empty.
        if labels_extra:
            ax.set_yticklabels([y + ('' if labels_extra[i] == '..' else '\n') for i, y in enumerate(labels)])
            for i, label in enumerate(ax.get_yticklabels()):
                if labels_extra[i] == '..': continue
                ax.text(0, i + 0.05, '\n' + labels_extra[i],
                        va='center', ha='right',
                        transform=label.get_transform(), color='#cafe0b')
        else: ax.set_yticklabels(labels)

        # Export to SVG
        fig.patch.set_visible(False) # hide the white background
        imgdata = io.StringIO()
        fig.savefig(imgdata, format='svg')
        plt.close() # otherwise it consumes a lot of memory in autoreload mode

        # Patch the rendered output: remove preable and hardcoded size
        imgdata = _patch_src.sub(_patch_dst, imgdata.getvalue())
        # Remove needless newlines and trailing whitespace in path data
        imgdata = _path_patch2_src.sub(_path_patch2_dst, _path_patch_src.sub(_path_patch_dst, imgdata))
        # Replace color codes with CSS classes
        for src, dst in _class_mapping: imgdata = src.sub(dst, imgdata)
        # Add titles for bars
        for i in range(len(value_sets)):
            for j in range(len(labels)):
                id = i*len(labels) + j
                if error_sets[i]: imgdata = imgdata.replace(
                    _bar_titles_src.format(mpl.rcParams['svg.hashsalt'], i, j),
                    _bar_titles_dst_error.format(mpl.rcParams['svg.hashsalt'], i, j, value_sets[i][j], error_sets[i][j], units))
                else: imgdata = imgdata.replace(
                    _bar_titles_src.format(mpl.rcParams['svg.hashsalt'], i, j),
                    _bar_titles_dst.format(mpl.rcParams['svg.hashsalt'], i, j, value_sets[i][j], units))

        container = nodes.container(**self.options)
        container['classes'] += ['m-plot']
        node = nodes.raw('', imgdata, format='html')
        container.append(node)
        return [container]

def new_page(*args, **kwargs):
    # The parameter is a number but "Support for setting an rcParam that
    # expects a str value to a non-str value is deprecated since 3.5 and
    # support will be removed two minor releases later", so we pass a string.
    mpl.rcParams['svg.hashsalt'] = '0'

def register_mcss(mcss_settings, hooks_pre_page, **kwargs):
    font = mcss_settings.get('M_PLOTS_FONT', 'Source Sans Pro')
    mpl.rcParams['font.family'] = font

    hooks_pre_page += [new_page]

    rst.directives.register_directive('plot', Plot)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    register_mcss(mcss_settings=pelicanobj.settings, hooks_pre_page=[])

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)
    signals.content_object_init.connect(new_page)



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

# This file is here only to make python unittest work, it's not needed
# otherwise



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#   Copyright © 2024 John Turner <7strbass@gmail.com>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import os.path

import docutils
from docutils.parsers import rst
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst import Directive, directives
import docutils.parsers.rst.directives.misc
from docutils import io, nodes, utils, statemachine

from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import TextLexer, BashSessionLexer, get_lexer_by_name

import logging

logger = logging.getLogger(__name__)

try:
    import ansilexer
except ImportError:
    # The above worked well on Pelican 4.2 and before, and also works with
    # other m.css tools like the Python doc generator. Pelican 4.5.0 changed to
    # "namespace plugins" and broke packaged plugins completely, 4.5.1 was
    # fixed to load namespaced plugins again, however the loading code is
    # different from 4.2 and thus anything from the root plugins/ directory
    # *isn't* in PATH anymore. Thus attempting to import those modules fails
    # and as a DIRTY hack I have to add the path back.
    #
    # TODO: Pelican 4.5+ treats everything that isn't in the pelican.plugins
    # namespace as "legacy plugins", which is unfortunate because then I
    # wouldn't be able to easily share the plugin code with other m.css tools
    # which don't (and shouldn't need to) care about Pelican at all. Allowing
    # 3rd party plugins without enforcing implicit assumptions on them (the
    # namespace, an unprefixed register() function...) would probably involve a
    # complex discussion with Pelican maintainers which I don't have the energy
    # for right now. Let's hope the "legacy plugins" codepath stays in for the
    # foreseeable future.
    import sys
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    import ansilexer

filters_pre = None
filters_post = None

def _highlight(code, language, options, *, is_block, filters=[]):
    # Use our own lexer for ANSI
    if language == 'ansi':
        lexer = ansilexer.AnsiLexer()
    else:
        try:
            lexer = get_lexer_by_name(language)
        except ValueError:
            logger.warning("No lexer found for language '{}', code highlighting disabled".format(language))
            lexer = TextLexer()

    if (isinstance(lexer, BashSessionLexer) or
        isinstance(lexer, ansilexer.AnsiLexer)):
        class_ = 'm-console'
    else:
        class_ = 'm-code'

    # Pygments wants the underscored option
    if 'hl-lines' in options:
        options['hl_lines'] = options['hl-lines']
        del options['hl-lines']

    if isinstance(lexer, ansilexer.AnsiLexer):
        formatter = ansilexer.HtmlAnsiFormatter(**options)
    else:
        formatter = HtmlFormatter(nowrap=True, **options)

    global filters_pre
    # First apply local pre filters, if any
    for filter in filters:
        f = filters_pre.get((lexer.name, filter))
        if f: code = f(code)
    # Then a global pre filter, if any
    f = filters_pre.get(lexer.name)
    if f: code = f(code)

    highlighted = highlight(code, lexer, formatter).rstrip()
    # Strip whitespace around if inline code, strip only trailing whitespace if
    # a block
    if not is_block: highlighted = highlighted.lstrip()

    global filters_post
    # First apply local post filters, if any
    for filter in filters:
        f = filters_post.get((lexer.name, filter))
        if f: highlighted = f(highlighted)
    # Then a global post filter, if any
    f = filters_post.get(lexer.name)
    if f: highlighted = f(highlighted)

    return class_, highlighted

class Code(Directive):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        'hl-lines': directives.unchanged,
        # Legacy alias to hl-lines (I hate underscores)
        'hl_lines': directives.unchanged,
        'class': directives.class_option,
        'filters': directives.unchanged
    }
    has_content = True

    def run(self):
        self.assert_has_content()

        set_classes(self.options)
        classes = []
        if 'classes' in self.options:
            classes += self.options['classes']
            del self.options['classes']

        # Legacy alias to hl-lines
        if 'hl_lines' in self.options:
            self.options['hl-lines'] = self.options['hl_lines']
            del self.options['hl_lines']

        filters = self.options.pop('filters', '').split()

        class_, highlighted = _highlight('\n'.join(self.content), self.arguments[0], self.options, is_block=True, filters=filters)
        classes += [class_]

        content = nodes.raw('', highlighted, format='html')
        pre = nodes.literal_block('', classes=classes)
        pre.append(content)
        return [pre]

class Include(docutils.parsers.rst.directives.misc.Include):
    option_spec = {
        'code': directives.unchanged,
        'tab-width': int,
        'start-line': int,
        'end-line': int,
        'start-after': directives.unchanged_required,
        'start-on': directives.unchanged_required,
        'end-before': directives.unchanged,
        'strip-prefix': directives.unchanged,
        'class': directives.class_option,
        'filters': directives.unchanged,
        'hl-lines': directives.unchanged
    }
    has_content = False

    def run(self):
        """
        Verbatim copy of docutils.parsers.rst.directives.misc.Include.run()
        that just calls to our Code instead of builtin CodeBlock, is without
        the rarely useful :encoding:, :literal: and :name: options and adds
        support for :start-on:, empty :end-before: and :strip-prefix:.
        """
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        source_dir = os.path.dirname(os.path.abspath(source))
        path = directives.path(self.arguments[0])
        if path.startswith('<') and path.endswith('>'):
            path = os.path.join(self.standard_include_path, path[1:-1])
        path = os.path.normpath(os.path.join(source_dir, path))
        path = utils.relative_path(None, path)
        e_handler=self.state.document.settings.input_encoding_error_handler
        tab_width = self.options.get(
            'tab-width', self.state.document.settings.tab_width)
        try:
            self.state.document.settings.record_dependencies.add(path)
            include_file = io.FileInput(source_path=path,
                                        error_handler=e_handler)
        except UnicodeEncodeError as error:
            raise self.severe('Problems with "%s" directive path:\n'
                              'Cannot encode input file path "%s" '
                              '(wrong locale?).' %
                              (self.name, SafeString(path)))
        except IOError as error:
            raise self.severe('Problems with "%s" directive path:\n%s.' %
                      (self.name, str(error)))
        startline = self.options.get('start-line', None)
        endline = self.options.get('end-line', None)
        try:
            if startline or (endline is not None):
                lines = include_file.readlines()
                rawtext = ''.join(lines[startline:endline])
            else:
                rawtext = include_file.read()
        except UnicodeError as error:
            raise self.severe('Problem with "%s" directive:\n%s' %
                              (self.name, str(error)))
        # start-after/end-before: no restrictions on newlines in match-text,
        # and no restrictions on matching inside lines vs. line boundaries
        after_text = self.options.get('start-after', None)
        if after_text:
            # skip content in rawtext before *and incl.* a matching text
            after_index = rawtext.find(after_text)
            if after_index < 0:
                raise self.severe('Problem with "start-after" option of "%s" '
                                  'directive:\nText not found.' % self.name)
            rawtext = rawtext[after_index + len(after_text):]
        # Compared to start-after, this includes the matched line
        on_text = self.options.get('start-on', None)
        if on_text:
            on_index = rawtext.find('\n' + on_text)
            if on_index < 0:
                raise self.severe('Problem with "start-on" option of "%s" '
                                  'directive:\nText not found.' % self.name)
            rawtext = rawtext[on_index:]
        # Compared to builtin include directive, the end-before can be empty,
        # in which case it simply matches the first empty line (which is
        # usually end of the code block)
        before_text = self.options.get('end-before', None)
        if before_text is not None:
            # skip content in rawtext after *and incl.* a matching text
            if before_text == '':
                before_index = rawtext.find('\n\n')
            else:
                before_index = rawtext.find(before_text)
            if before_index < 0:
                raise self.severe('Problem with "end-before" option of "%s" '
                                  'directive:\nText not found.' % self.name)
            rawtext = rawtext[:before_index]

        include_lines = statemachine.string2lines(rawtext, tab_width,
                                                  convert_whitespace=True)

        # Strip a common prefix from all lines. Useful for example when
        # including a reST snippet that's embedded in a comment, or cutting
        # away excessive indentation. Can be wrapped in quotes in order to
        # avoid trailing whitespace in reST markup.
        if 'strip-prefix' in self.options and self.options['strip-prefix']:
            prefix = self.options['strip-prefix']
            if prefix[0] == prefix[-1] and prefix[0] in ['\'', '"']:
                prefix = prefix[1:-1]
            for i, line in enumerate(include_lines):
                if line.startswith(prefix): include_lines[i] = line[len(prefix):]
                # Strip the prefix also if the line is just the prefix alone,
                # with trailing whitespace removed
                elif line.rstrip() == prefix.rstrip(): include_lines[i] = ''

        if 'code' in self.options:
            self.options['source'] = path
            # Don't convert tabs to spaces, if `tab_width` is negative:
            if tab_width < 0:
                include_lines = rawtext.splitlines()
            codeblock = Code(self.name,
                                  [self.options.pop('code')], # arguments
                                  self.options,
                                  include_lines, # content
                                  self.lineno,
                                  self.content_offset,
                                  self.block_text,
                                  self.state,
                                  self.state_machine)
            return codeblock.run()
        self.state_machine.insert_input(include_lines, path)
        return []

def code(role, rawtext, text, lineno, inliner, options={}, content=[]):
    # In order to properly preserve backslashes (well, and backticks)
    text = rawtext[rawtext.find('`') + 1:rawtext.rfind('`')]

    set_classes(options)
    classes = []
    if 'classes' in options:
        classes += options['classes']
        del options['classes']

    # If language is not specified, render a simple literal
    if not 'language' in options:
        content = nodes.raw('', utils.unescape(text), format='html')
        node = nodes.literal(rawtext, '', **options)
        node.append(content)
        return [node], []

    language = options['language']
    del options['language']
    # Not sure why language is duplicated in classes?
    if language in classes: classes.remove(language)

    filters = options.pop('filters', '').split()

    class_, highlighted = _highlight(utils.unescape(text), language, options, is_block=False, filters=filters)
    classes += [class_]

    content = nodes.raw('', highlighted, format='html')
    node = nodes.literal(rawtext, '', classes=classes, **options)
    node.append(content)
    return [node], []

code.options = {'class': directives.class_option,
                'language': directives.unchanged,
                'filters': directives.unchanged}

def register_mcss(mcss_settings, **kwargs):
    rst.directives.register_directive('code', Code)
    rst.directives.register_directive('include', Include)
    rst.roles.register_canonical_role('code', code)

    # These two are builtin aliases to .. code:: in docutils:
    # https://github.com/docutils-mirror/docutils/blob/e88c5fb08d5cdfa8b4ac1020dd6f7177778d5990/docutils/parsers/rst/languages/en.py#L22-L24
    # Since a lot of existing markup (especially coming from Sphinx) uses
    # .. code-block:: and since there's no reason for .. code-block:: /
    # .. sourcecode:: to behave like unpatched docutils, let's add those too:
    rst.directives.register_directive('code-block', Code)
    rst.directives.register_directive('sourcecode', Code)

    global filters_pre, filters_post
    filters_pre = mcss_settings.get('M_CODE_FILTERS_PRE', {})
    filters_post = mcss_settings.get('M_CODE_FILTERS_POST', {})

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    settings = {}
    for key in ['M_CODE_FILTERS_PRE', 'M_CODE_FILTERS_POST']:
        if key in pelicanobj.settings: settings[key] = pelicanobj.settings[key]

    register_mcss(mcss_settings=settings)

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Loosely based on https://github.com/Nitron/pelican-alias,
#   copyright © 2013 Christopher Williams
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import os
import logging
from urllib.parse import urljoin

logger = logging.getLogger(__name__)

class AliasGenerator:
    def __init__(self, context, settings, path, theme, output_path, *args):
        self.output_path = output_path
        self.context = context
        self.siteurl = settings['SITEURL']

    # Keep consistent with m.htmlsanity
    def format_siteurl(self, url):
        return urljoin(self.siteurl + ('/' if not self.siteurl.endswith('/') else ''), url)

    def generate_output(self, writer):
        for page in self.context['pages'] + self.context['articles']:
            aliases = page.metadata.get('alias', '').strip()
            if not aliases: continue
            for alias in aliases.split('\n'):
                alias = alias.strip()

                # Currently only supporting paths starting with /
                if not alias.startswith('/'): assert False
                path = os.path.join(self.output_path, alias[1:])

                # If path ends with /, write it into index.html
                directory, filename = os.path.split(path)
                if not filename: filename = 'index.html'

                alias_file = os.path.join(directory, filename)
                alias_target = self.format_siteurl(page.url)
                logger.info('m.alias: creating alias {} -> {}'.format(alias_file[len(self.output_path):], alias_target))

                # Write the redirect file
                os.makedirs(directory, exist_ok=True)
                with open(alias_file, 'w') as f:
                    f.write("""<!DOCTYPE html><html><head><meta http-equiv="refresh" content="0;url={}" /></head></html>\n""".format(alias_target))

def register_mcss(**kwargs):
    assert not "This plugin is Pelican-only" # pragma: no cover

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_get_generators(generators): return AliasGenerator

def register(): # for Pelican
    from pelican import signals

    signals.get_generators.connect(_pelican_get_generators)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and classes for working with emoji.
"""

import json
from typing import Collection, Union

from . import paths
from .utils import *


def update_database_file():
    # get the raw emoji db from github
    EMOJI_URI = r'https://api.github.com/emojis'
    print(rf"Downloading {EMOJI_URI}")
    raw_emoji = json.loads(download_text(EMOJI_URI))

    # convert into our internal Emoji tuple
    emoji = dict()
    RX_GITHUB_EMOJI_URI = re.compile(r".+unicode/([0-9a-f]+(?:-[0-9a-f]+)?)[.]png.*", re.I)
    for key, uri in raw_emoji.items():
        m = RX_GITHUB_EMOJI_URI.fullmatch(uri)
        if m:
            cps = [int(cp, 16) for cp in m[1].split('-')]
            if len(cps) == 1:
                cps = cps[0]
            emoji[key.lower().replace(r'-', r'_')] = (cps, uri)

    # serialize
    path = Path(paths.GENERATED, r'emoji.json')
    print(rf'Writing {path}')
    with open(path, 'w', encoding='utf-8', newline='\n') as f:
        f.write(json.dumps(emoji, sort_keys=True, indent=4))


class Emoji(object):
    def __init__(self, key: str, codepoints: Union[int, Collection[int]], uri: str):
        self.key = str(key)
        self.codepoints = [int(cp) for cp in coerce_collection(codepoints)]
        self.codepoints.sort()
        self.codepoints = tuple(self.codepoints)
        self.uri = str(uri)

    def __str__(self) -> str:
        s = ''
        for cp in self.codepoints:
            s += rf'&#x{cp:X};'
        return rf'{s}&#xFE0F;'


class Database(object):
    def __init__(self):
        path = Path(paths.GENERATED, r'emoji.json')
        assert_existing_file(path)
        emoji = json.loads(read_all_text_from_file(path))

        # load by key
        self.__by_key = dict()
        for key, vals in emoji.items():
            self.__by_key[key] = Emoji(key, vals[0], vals[1])

        # load by codepoint
        self.__by_codepoint = dict()
        for key, e in self.__by_key.items():
            if len(e.codepoints) == 1:
                self.__by_codepoint[e.codepoints[0]] = e

        # apply aliases
        ALIASES = (
            ('sundae', 'ice_cream'),  #
            ('info', 'information_source'),
            ('man_in_tuxedo', 'person_in_tuxedo'),
            ('bride_with_veil', 'person_with_veil'),
        )
        for alias, key in ALIASES:
            if key in self.__by_key and alias not in self.__by_key:
                self.__by_key[alias] = self.__by_key[key]

    def __contains__(self, key: Union[int, str]) -> bool:
        assert key is not None
        if isinstance(key, int):
            return key in self.__by_codepoint
        else:
            return key.lower().replace(r'-', r'_') in self.__by_key

    def __getitem__(self, key: Union[int, str]) -> Emoji:
        assert key is not None
        if isinstance(key, int):
            if key in self.__by_codepoint:
                return self.__by_codepoint[key]
        else:
            key = key.lower().replace(r'-', r'_')
            if key in self.__by_key:
                return self.__by_key[key]
        return None


__all__ = ['update_database_file', 'Emoji', 'Database']



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import datetime

from schema import And, Optional, Or, Schema, SchemaError, Use

py2toml = {
    str: r'string',
    list: r'array',
    dict: r'table',
    int: r'integer',
    float: r'float',
    bool: r'boolean',
    datetime.date: r'date',
    datetime.time: r'time',
    datetime.datetime: r'date-time',
}


def FixedArrayOf(typ, length, name=''):
    global py2toml
    return And(
        [typ],
        lambda v: len(v) == length,
        error=rf'{name + ": " if name else ""}expected array of {length} {py2toml[typ]}{"s" if length != 1 else ""}',
    )


def ValueOrArray(typ, name='', length=None):
    global py2toml
    inner = None
    if length is None:
        inner = Or(typ, [typ], error=rf'{name + ": " if name else ""}expected {py2toml[typ]} or array of {py2toml[typ]}s')
    else:
        err = rf'{name + ": " if name else ""}expected {py2toml[typ]} or array of {length} {py2toml[typ]}{"s" if length != 1 else ""}'
        inner = And(Or(typ, [typ], error=err), lambda v: not isinstance(v, list) or len(v) == length, error=err)
    return And(inner, Use(lambda x: x if isinstance(x, list) else [x]))


def Stripped(typ, allow_empty=True, name=''):
    if not name:
        name = 'value'
    return And(
        And(typ, Use(lambda x: x.strip())),
        (lambda x: True) if allow_empty else (lambda x: len(x) > 0),
        error=rf'{name} cannot be blank',
    )


context_stack = list()


class SchemaContext(object):
    def __init__(self, val: str):
        self.__val = str(val) if val is not None else None

    def __enter__(self):
        global context_stack
        if self.__val:
            context_stack.append(self.__val)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.__val and exc_type is None:
            global context_stack
            context_stack.pop()


def current_schema_context() -> str:
    global context_stack
    ctx = ''
    for val in context_stack:
        ctx += rf'{val}: '
    return ctx


__all__ = [
    'And',
    'Optional',
    'Or',
    'Schema',
    'Use',
    'FixedArrayOf',
    'ValueOrArray',
    'Stripped',
    'SchemaContext',
    'current_schema_context',
    'SchemaError',
]



#!/usr/bin/env python

#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import os
import sys

# So it's possible to do `./plugins/m/sphinx.py somefile.inv` without it all
# blowing up on the earliest `import math` as there's a math.py file in this
# directory. Alternatively, doing `python -m m.sphinx` would work too, but who
# would remember to do that.
if __name__ == '__main__': # pragma: no cover
    sys.path.remove(os.path.realpath(os.path.dirname(__file__)))

import argparse
import logging
import re
from types import SimpleNamespace as Empty
from typing import Dict, List, Optional
from urllib.parse import urljoin
import zlib

import docutils
from docutils import nodes, utils
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst.states import Inliner

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
import m.htmlsanity

# All those initialized in register() or register_mcss()
current_referer_path = None
page_ref_prefixes = None
current_param_names = None
module_doc_output = None
class_doc_output = None
enum_value_doc_output = None
enum_doc_output = None
function_doc_output = None
property_doc_output = None
data_doc_output = None
inventory_filename = None

# List option (allowing multiple arguments). See _docutils_assemble_option_dict
# in python.py for details.
def directives_unchanged_list(argument):
    return [directives.unchanged(argument)]

class PyModule(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged,
                   'data': directives_unchanged_list}

    def run(self):
        # Check that data are parsed properly, turn them into a dict. This will
        # blow up if the data name is not specified.
        properties = {}
        data = {}
        for name, summary in self.options.get('data', []):
            if name in data: raise KeyError("duplicate data {}".format(name))
            data[name] = summary

        output = module_doc_output.setdefault(self.arguments[0], {})
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if self.content:
            output['content'] = '\n'.join(self.content)

        for name, summary in data.items():
            data_doc_output.setdefault('{}.{}'.format(self.arguments[0], name), {})['summary'] = summary

        return []

class PyClass(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged,
                   'property': directives_unchanged_list,
                   'data': directives_unchanged_list}

    def run(self):
        # Check that properties and data are parsed properly, turn them into
        # dicts. This will blow up if the property / data name is not specified.
        properties = {}
        for name, summary in self.options.get('property', []):
            if name in properties: raise KeyError("duplicate property {}".format(name))
            properties[name] = summary
        data = {}
        for name, summary in self.options.get('data', []):
            if name in data: raise KeyError("duplicate data {}".format(name))
            data[name] = summary

        output = class_doc_output.setdefault(self.arguments[0], {})
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if self.content:
            output['content'] = '\n'.join(self.content)

        for name, summary in properties.items():
            property_doc_output.setdefault('{}.{}'.format(self.arguments[0], name), {})['summary'] = summary
        for name, summary in data.items():
            data_doc_output.setdefault('{}.{}'.format(self.arguments[0], name), {})['summary'] = summary

        return []

class PyEnum(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged,
                   'value': directives_unchanged_list}

    def run(self):
        # Check that values are parsed properly, turn them into a dict. This
        # will blow up if the param name is not specified.
        values = {}
        for name, summary in self.options.get('value', []):
            if name in values: raise KeyError("duplicate param {}".format(name))
            values[name] = summary

        output = enum_doc_output.setdefault(self.arguments[0], {})
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if self.content:
            output['content'] = '\n'.join(self.content)

        for name, content in values.items():
            enum_value_doc_output.setdefault('{}.{}'.format(self.arguments[0], name), {})['content'] = content

        return []

class PyEnumValue(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1

    def run(self):
        output = enum_value_doc_output.setdefault(self.arguments[0], {})
        if self.content:
            output['content'] = '\n'.join(self.content)
        return []

class PyFunction(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged,
                   'param': directives_unchanged_list,
                   'raise': directives_unchanged_list,
                   'return': directives.unchanged}

    def run(self):
        # Check that params are parsed properly, turn them into a dict. This
        # will blow up if the param name is not specified.
        params = {}
        for name, content in self.options.get('param', []):
            if name in params: raise KeyError("duplicate param {}".format(name))
            params[name] = content
        # Check that exceptions are parsed properly. This will blow up if the
        # exception name is not specified. Unlike function params not turning
        # these into a dict since a single type can be raised for multiple
        # reasons.
        raises = []
        for name, content in self.options.get('raise', []):
            raises += [(name, content)]

        output = function_doc_output.setdefault(self.arguments[0], {})
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if params:
            output['params'] = params
        if raises:
            output['raise'] = raises
        if self.options.get('return'):
            output['return'] = self.options['return']
        if self.content:
            output['content'] = '\n'.join(self.content)
        return []

class PyProperty(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged,
                   'raise': directives_unchanged_list}

    def run(self):
        # Check that exceptions are parsed properly. This will blow up if the
        # exception name is not specified. Unlike function params not turning
        # these into a dict since a single type can be raised for multiple
        # reasons.
        raises = []
        for name, content in self.options.get('raise', []):
            raises += [(name, content)]

        output = property_doc_output.setdefault(self.arguments[0], {})
        if raises:
            output['raise'] = raises
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if self.content:
            output['content'] = '\n'.join(self.content)
        return []

class PyData(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'summary': directives.unchanged}

    def run(self):
        output = data_doc_output.setdefault(self.arguments[0], {})
        if self.options.get('summary'):
            output['summary'] = self.options['summary']
        if self.content:
            output['content'] = '\n'.join(self.content)
        return []

# Modified from abbr / gh / gl / ... to add support for queries and hashes
link_regexp = re.compile(r'(?P<title>.*) <(?P<link>[^?#]+)(?P<hash>[?#].+)?>')

def parse_link(text):
    link = utils.unescape(text)
    m = link_regexp.match(link)
    if m:
        title, link, hash = m.group('title', 'link', 'hash')
        if not hash: hash = '' # it's None otherwise
    else:
        title, hash = '', ''

    return title, link, hash

intersphinx_inventory = {}
intersphinx_name_prefixes = []

# Basically a copy of sphinx.util.inventory.InventoryFile.load_v2. There's no
# documentation for this, it seems.
def parse_intersphinx_inventory(file, base_url, inventory, css_classes):
    # Parse the header, uncompressed
    inventory_version = file.readline().rstrip()
    if inventory_version != b'# Sphinx inventory version 2':
        raise ValueError("Unsupported inventory version header: {}".format(inventory_version)) # pragma: no cover
    # those two are not used at the moment, just for completeness
    project = file.readline().rstrip()[11:]
    version = file.readline().rstrip()[11:]
    line = file.readline()
    if b'zlib' not in line:
        raise ValueError("invalid inventory header (not compressed): {}".format(line)) # pragma: no cover

    # Decompress the rest. Again mostly a copy of the sphinx code.
    for line in zlib.decompress(file.read()).decode('utf-8').splitlines():
        m = re.match(r'(?x)(.+?)\s+(\S*:\S*)\s+(-?\d+)\s+(\S+)\s+(.*)',
                         line.rstrip())
        if not m: # pragma: no cover
            print("wait what is this line?! {}".format(line))
            continue
        # What the F is prio for
        name, type, prio, location, title = m.groups()

        # What is this?!
        if location.endswith('$'): location = location[:-1] + name

        # The original code `continue`s in this case. I'm asserting. Fix your
        # docs.
        assert not(type == 'py:module' and type in inventory and name in inventory[type]), "Well dang, we hit that bug in 1.1 that I didn't want to work around" # pragma: no cover

        # Prepend base URL and add to the inventory
        inventory.setdefault(type, {})[name] = (urljoin(base_url, location), title, css_classes)

def parse_intersphinx_inventories(input, inventories):
    global intersphinx_inventory, intersphinx_name_prefixes
    intersphinx_inventory = {}
    intersphinx_name_prefixes = ['']

    for f in inventories:
        inventory, base_url = f[:2]
        prefixes = f[2] if len(f) > 2 else []
        css_classes = f[3] if len(f) > 3 else []

        intersphinx_name_prefixes += prefixes
        with open(os.path.join(input, inventory), 'rb') as file:
            parse_intersphinx_inventory(file, base_url, intersphinx_inventory, css_classes)

# Matches e.g. py:function in py:function:open
_type_prefix_re = re.compile(r'([a-z0-9]{,3}:[a-z0-9]{3,}):')
_function_types = ['py:function', 'py:classmethod', 'py:staticmethod', 'py:method', 'c:function']

def ref(name, rawtext, text, lineno, inliner: Inliner, options={}, content=[]):
    title, target, hash = parse_link(text)

    # Otherwise adding classes to the options behaves globally (uh?)
    _options = dict(options)
    set_classes(_options)
    # Avoid assert on adding to undefined member later
    if 'classes' not in _options: _options['classes'] = []

    # If we're in a page and there's no page module scope yet, look if there
    # are :py:module: page metadata we could use for a prefix
    global current_referer_path, page_ref_prefixes
    if current_referer_path[-1][0].name == 'PAGE' and page_ref_prefixes is None:
        # Since we're in the middle of parse, the nodes.docinfo is not present
        # yet (it's produced by the frontmatter.DocInfo transform that's run
        # after the parsing ends), so we look in field lists instead.
        # TODO: DocInfo picks only the first ever field list happening right
        #   after a title and we should do the same to avoid picking up options
        #   later in the page. There the transform depends on DocTitle being
        #   ran already, so it would need to be more complex here. See
        #   docutils.transforms.frontmatter.DocInfo.apply() for details.
        for docinfo in inliner.document.traverse(docutils.nodes.field_list):
            for element in docinfo.children:
                if element.tagname != 'field': continue
                name_elem, body_elem = element.children
                if name_elem.astext() == 'ref-prefix':
                    page_ref_prefixes = [line.strip() + '.' for line in body_elem.astext().splitlines() if line.strip()]

        # If we didn't find any, set it to an empty list (not None), so this is
        # not traversed again next time
        if not page_ref_prefixes: page_ref_prefixes = []

    # Add prefixes of the referer path to the global prefix list. Make an empty
    # prefix first so :ref:`open()` always reliably links to the builtin
    # instead whatever `foo.bar.open()` that's currently in scope. After that
    # the names "closest" to the referer have the biggest priority with shorter
    # referer_path prefixes after, and name prefixes from M_SPHINX_INVENTORIES
    # last.
    global intersphinx_inventory, intersphinx_name_prefixes
    referer_path = current_referer_path[-1][1] if current_referer_path else []
    prefixes = [''] + ['.'.join(referer_path[:len(referer_path) - i]) + '.' for i, _ in enumerate(referer_path)] + (page_ref_prefixes if page_ref_prefixes else []) + intersphinx_name_prefixes
    for prefix in prefixes:
        found = None

        # If the target is prefixed with a type, try looking up that type
        # directly. The implicit link title is then without the type.
        m = _type_prefix_re.match(target)
        if m:
            type = m.group(1)
            prefixed = prefix + target[len(type) + 1:]
            # ALlow trailing () on functions here as well
            if prefixed.endswith('()') and type in _function_types:
                prefixed = prefixed[:-2]
            if type in intersphinx_inventory and prefixed in intersphinx_inventory[type]:
                target = target[len(type) + 1:]
                found = type, intersphinx_inventory[m.group(1)][prefixed]

        prefixed = prefix + target

        # If the target looks like a function, look only in functions and strip
        # the trailing () as the inventory doesn't have that
        if not found and prefixed.endswith('()'):
            prefixed = prefixed[:-2]
            for type in _function_types:
                if type in intersphinx_inventory and prefixed in intersphinx_inventory[type]:
                    found = type, intersphinx_inventory[type][prefixed]
                    break

        # Iterate through whitelisted types otherwise. Skipping
        # 'std:pdbcommand', 'std:cmdoption', 'std:term', 'std:label',
        # 'std:opcode', 'std:envvar', 'std:token', 'std:doc', 'std:2to3fixer'
        # and unknown domains such as c++ for now as I'm unsure about potential
        # name clashes.
        if not found:
            for type in [
                'py:exception', 'py:attribute', 'py:method', 'py:data', 'py:module', 'py:function', 'py:class', 'py:classmethod', 'py:staticmethod',
                'c:var', 'c:type', 'c:function', 'c:member', 'c:macro',
                # TODO: those apparently don't exist:
                'py:enum', 'py:enumvalue'
            ]:
                if type in intersphinx_inventory and prefixed in intersphinx_inventory[type]:
                    found = type, intersphinx_inventory[type][prefixed]

        if found:
            url, link_title, css_classes = found[1]
            if title:
                use_title = title
            elif link_title != '-':
                use_title = link_title
            else:
                use_title = target
                # Add () to function refs
                if found[0] in _function_types and not target.endswith('()'):
                    use_title += '()'

            _options['classes'] += css_classes
            node = nodes.reference(rawtext, use_title, refuri=url + hash, **_options)
            return [node], []

    if title:
        logging.warning("Sphinx symbol `{}` not found, rendering just link title".format(target))
        node = nodes.inline(rawtext, title, **_options)
    else:
        logging.warning("Sphinx symbol `{}` not found, rendering as monospace".format(target))
        node = nodes.literal(rawtext, target, **_options)
    return [node], []

def scope_enter(type, path, param_names=None, **kwargs):
    global current_referer_path, current_param_names
    current_referer_path += [(type, path)]
    current_param_names = param_names

    # If we're in a page, reset page_ref_prefixes so next time :ref: needs it,
    # it will look it up instead of using a stale version
    global page_ref_prefixes
    if type.name == 'PAGE': page_ref_prefixes = None

def scope_exit(type, path, **kwargs):
    global current_referer_path, current_param_names
    assert current_referer_path[-1] == (type, path), "%s %s" % (current_referer_path, path)
    current_referer_path = current_referer_path[:-1]
    current_param_names = None

def check_scope_stack_empty(**kwargs):
    global current_referer_path
    assert not current_referer_path

def p(name, rawtext, text, lineno, inliner: Inliner, options={}, content=[]):
    global current_referer_path, current_param_names
    if not current_param_names:
        logging.warning("can't reference parameter %s outside of a function scope", text)
    elif text not in current_param_names:
        logging.warning("parameter %s not found in %s(%s) function signature", text, '.'.join(current_referer_path[-1][1]), ', '.join(current_param_names))

    node = nodes.literal(rawtext, text, **options)
    return [node], []

def consume_docstring(type, path: List[str], signature: Optional[str], doc: str) -> str:
    # Create the directive header based on type
    if type.name == 'MODULE':
        source = '.. py:module:: '
        doc_output = module_doc_output
    elif type.name == 'CLASS':
        source = '.. py:class:: '
        doc_output = class_doc_output
    elif type.name == 'ENUM':
        source = '.. py:enum:: '
        doc_output = enum_doc_output
    elif type.name == 'ENUM_VALUE':
        source = '.. py:enumvalue:: '
        doc_output = enum_value_doc_output
    elif type.name in ['FUNCTION', 'OVERLOADED_FUNCTION']:
        source = '.. py:function:: '
        doc_output = function_doc_output
    elif type.name == 'PROPERTY':
        source = '.. py:property:: '
        doc_output = property_doc_output
    else: # pragma: no cover
        # Data don't have docstrings, you silly
        assert type.name != 'DATA'
        # Ignore unknown types, pass the docs through
        return doc

    # Add path and signature to the header
    path_signature_str = '.'.join(path) + (signature if signature else '')
    source += path_signature_str + '\n'

    # Everything except enum values has a separate summary. Assuming first
    # paragraph is a summary, turn it into a :summary: directive option with
    # successive lines indented.
    if type.name != 'ENUM_VALUE':
        summary, _, doc = doc.partition('\n\n')
        source += '    :summary: {}\n'.format(summary.replace('\n', '\n        '))

    # The next paragraph could be option list. If that's so, indent those as
    # well, append
    if doc.startswith(':'):
        options, _, doc = doc.partition('\n\n')
        source += '    {}\n\n'.format(options.replace('\n', '\n    '))
    else:
        source += '\n'

    # The rest (if any) is content. Indent as well.
    source += '    {}\n'.format(doc.replace('\n', '\n    '))

    # Unleash docutils on this piece. It will call into the proper directive
    # and do the thing. Ignore the output as there shouldn't be anything left.
    pub = docutils.core.Publisher(
        writer=m.htmlsanity.SaneHtmlWriter(),
        source_class=docutils.io.StringInput,
        destination_class=docutils.io.StringOutput)
    pub.set_components('standalone', 'restructuredtext', 'html')
    pub.writer.translator_class = m.htmlsanity.SaneHtmlTranslator
    pub.process_programmatic_settings(None, m.htmlsanity.docutils_settings, None)
    # Docutils uses a deprecated U mode for opening files, so instead of
    # monkey-patching docutils.io.FileInput to not do that (like Pelican does),
    # I just read the thing myself.
    # TODO it *somehow* needs to supply the original docstring filename and
    # line range to it for better error reporting, this is too awful
    pub.set_source(source=source)
    pub.publish()

def merge_inventories(name_map, **kwargs):
    global intersphinx_inventory

    # Create inventory entries from the name_map
    internal_inventory = {}
    for path_str, entry in name_map.items():
        EntryType = type(entry.type) # so we don't need to import the enum
        if entry.type == EntryType.MODULE:
            type_string = 'py:module'
        elif entry.type == EntryType.CLASS:
            type_string = 'py:class'
        elif entry.type == EntryType.FUNCTION:
            # TODO: properly distinguish between 'py:function',
            # 'py:classmethod', 'py:staticmethod', 'py:method'
            type_string = 'py:function'
        elif entry.type == EntryType.OVERLOADED_FUNCTION:
            # TODO: what about the other overloads?
            type_string = 'py:function'
        elif entry.type == EntryType.PROPERTY:
            # datetime.date.year is decorated with @property and listed as a
            # py:attribute, so that's probably it
            type_string = 'py:attribute'
        elif entry.type == EntryType.ENUM:
            type_string = 'py:enum' # this doesn't exist in Sphinx
        elif entry.type == EntryType.ENUM_VALUE:
            type_string = 'py:enumvalue' # these don't exist in Sphinx
        elif entry.type == EntryType.DATA:
            type_string = 'py:data'
        elif entry.type == EntryType.PAGE:
            type_string = 'std:doc'
        elif entry.type == EntryType.SPECIAL:
            # TODO: this will cause duplicates when multiple m.css projects
            #   gets together, solve better
            type_string = 'std:special'
        else: # pragma: no cover
            assert False

        # Mark those with m-doc (as internal)
        internal_inventory.setdefault(type_string, {})[path_str] = (entry.url, '-', ['m-doc'])

    # Add class / enum / enum value inventory entries to the name map for type
    # cross-linking
    for type_, type_string in [
        # TODO: this will blow up if the above loop is never entered (which is
        # unlikely) as EntryType is defined there
        (EntryType.CLASS, 'py:class'),
        # Otherwise we can't link to standard exceptions from :raise:
        (EntryType.CLASS, 'py:exception'), # TODO: special type for these?
        (EntryType.DATA, 'py:data'), # typing.Tuple or typing.Any is data
        # Those are custom to m.css, not in Sphinx
        (EntryType.ENUM, 'py:enum'),
        (EntryType.ENUM_VALUE, 'py:enumvalue'),
    ]:
        if type_string in intersphinx_inventory:
            for path, value in intersphinx_inventory[type_string].items():
                url, _, css_classes = value
                entry = Empty()
                entry.type = type_
                entry.object = None
                entry.path = path.split('.')
                entry.css_classes = css_classes
                entry.url = url
                name_map[path] = entry

    # Add stuff from the name map to our inventory
    for type_, data_internal in internal_inventory.items():
        data = intersphinx_inventory.setdefault(type_, {})
        for path, value in data_internal.items():
            # Ignore duplicate things (such as `index` etc.)
            # TODO: solve better
            if path in data: continue
            data[path] = value

    # Save the internal inventory, if requested. Again basically a copy of
    # sphinx.util.inventory.InventoryFile.dump().
    if inventory_filename:
        with open(os.path.join(inventory_filename), 'wb') as f:
            # Header
            # TODO: user-defined project/version
            f.write(b'# Sphinx inventory version 2\n'
                    b'# Project: X\n'
                    b'# Version: 0\n'
                    b'# The remainder of this file is compressed using zlib.\n')

            # Body. Sorting so it's in a reproducible order for testing.
            compressor = zlib.compressobj(9)
            for type_, data in sorted(internal_inventory.items()):
                for path, value in data.items():
                    url, title, css_classes = value
                    # The type has to contain a colon. Wtf is the 2?
                    assert ':' in type_
                    f.write(compressor.compress('{} {} 2 {} {}\n'.format(path, type_, url, title).encode('utf-8')))
            f.write(compressor.flush())

def register_mcss(mcss_settings, module_doc_contents, class_doc_contents, enum_doc_contents, enum_value_doc_contents, function_doc_contents, property_doc_contents, data_doc_contents, hooks_post_crawl, hooks_pre_scope, hooks_post_scope, hooks_docstring, hooks_post_run, **kwargs):
    global current_referer_path, module_doc_output, class_doc_output, enum_doc_output, enum_value_doc_output, function_doc_output, property_doc_output, data_doc_output, inventory_filename
    current_referer_path = []
    module_doc_output = module_doc_contents
    class_doc_output = class_doc_contents
    enum_doc_output = enum_doc_contents
    enum_value_doc_output = enum_value_doc_contents
    function_doc_output = function_doc_contents
    property_doc_output = property_doc_contents
    data_doc_output = data_doc_contents
    inventory_filename = os.path.join(mcss_settings['OUTPUT'], mcss_settings['M_SPHINX_INVENTORY_OUTPUT']) if 'M_SPHINX_INVENTORY_OUTPUT' in mcss_settings else None

    parse_intersphinx_inventories(input=mcss_settings['INPUT'],
         inventories=mcss_settings.get('M_SPHINX_INVENTORIES', []))

    rst.directives.register_directive('py:module', PyModule)
    rst.directives.register_directive('py:class', PyClass)
    rst.directives.register_directive('py:enum', PyEnum)
    rst.directives.register_directive('py:enumvalue', PyEnumValue)
    rst.directives.register_directive('py:function', PyFunction)
    rst.directives.register_directive('py:property', PyProperty)
    rst.directives.register_directive('py:data', PyData)

    rst.roles.register_local_role('ref', ref)
    rst.roles.register_local_role('p', p)

    hooks_pre_scope += [scope_enter]
    hooks_post_scope += [scope_exit]
    if mcss_settings.get('M_SPHINX_PARSE_DOCSTRINGS', False):
        hooks_docstring += [consume_docstring]
    hooks_post_crawl += [merge_inventories]
    # Just a sanity check
    hooks_post_run += [check_scope_stack_empty]

def _pelican_new_page(generator):
    # Set a dummy page referrer path so :ref-prefixes: works in Pelican as well
    # TODO: any chance this could be made non-crappy?
    global current_referer_path
    assert not current_referer_path or len(current_referer_path) == 1 and current_referer_path[0][0].name == 'PAGE'
    type = Empty() # We don't have the EntryType enum, so fake it
    type.name = 'PAGE'
    current_referer_path = [(type, '')]

def _pelican_configure(pelicanobj):

    # For backwards compatibility, the input directory is pelican's CWD
    parse_intersphinx_inventories(input=os.getcwd(),
         inventories=pelicanobj.settings.get('M_SPHINX_INVENTORIES', []))

def register(): # for Pelican
    from pelican import signals

    rst.roles.register_local_role('ref', ref)

    signals.initialized.connect(_pelican_configure)
    signals.article_generator_preread.connect(_pelican_new_page)
    signals.page_generator_preread.connect(_pelican_new_page)

def pretty_print_intersphinx_inventory(file):
    return ''.join([
        # Sphinx inventory version 2
        file.readline().decode('utf-8'),
        # Project and version
        file.readline().decode('utf-8'),
        file.readline().decode('utf-8'),
        # Zlib compression line
        file.readline().decode('utf-8'),
        # The rest, zlib-compressed
        zlib.decompress(file.read()).decode('utf-8')
    ])

if __name__ == '__main__': # pragma: no cover
    parser = argparse.ArgumentParser()
    parser.add_argument('file', help="inventory file to print")
    parser.add_argument('--raw', help="show raw content", action='store_true')
    parser.add_argument('--types', help="list all type keys", action='store_true')
    parser.add_argument('--repack', help="repack raw content back into a compressed file", action='store_true')
    args = parser.parse_args()

    if args.types:
        with open(args.file, 'rb') as f:
            inventory = {}
            parse_intersphinx_inventory(f, '', inventory, [])
            print(inventory.keys())
    elif args.repack:
        with open(args.file, 'rb') as f:
            # Sphinx inventory version 2
            sys.stdout.buffer.write(f.readline())
            # Project and version
            sys.stdout.buffer.write(f.readline())
            sys.stdout.buffer.write(f.readline())
            # Zlib compression line
            sys.stdout.buffer.write(f.readline())
            # The rest, zlib-compressed
            sys.stdout.buffer.write(zlib.compress(f.read()))
    else:
        with open(args.file, 'rb') as f:
            print(pretty_print_intersphinx_inventory(f))



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#   Copyright © 2020 Blair Conrad <blair@blairconrad.com>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re

from pygments.lexer import RegexLexer
from pygments.formatters import HtmlFormatter
from pygments.token import *

# Support ANSI SGR codes in input, styling the output. Code definitions are
# taken from http://man7.org/linux/man-pages/man4/console_codes.4.html, which
# appears in part below, in case it disappears:
#
# ECMA-48 Set Graphics Rendition
#
#    The ECMA-48 SGR sequence ESC [ parameters m sets display attributes.
#    Several attributes can be set in the same sequence, separated by
#    semicolons.  An empty parameter (between semicolons or string
#    initiator or terminator) is interpreted as a zero.
#   param     result
#    0         reset all attributes to their defaults
#    1         set bold
#    2         set half-bright (simulated with color on a color display)
#    4         set underscore (simulated with color on a color display)
#              (the colors used to simulate dim or underline are set
#              using ESC ] ...)
#    5         set blink
#    7         set reverse video
#    10        reset selected mapping, display control flag, and toggle
#              meta flag (ECMA-48 says "primary font").
#    11        select null mapping, set display control flag, reset
#              toggle meta flag (ECMA-48 says "first alternate font").
#    12        select null mapping, set display control flag, set toggle
#              meta flag (ECMA-48 says "second alternate font").  The
#              toggle meta flag causes the high bit of a byte to be
#              toggled before the mapping table translation is done.
#    21        set underline; before Linux 4.17, this value set normal
#              intensity (as is done in many other terminals)
#    22        set normal intensity
#    24        underline off
#    25        blink off
#    27        reverse video off
#    30        set black foreground
#    31        set red foreground
#    32        set green foreground
#    33        set brown foreground
#    34        set blue foreground
#    35        set magenta foreground
#    36        set cyan foreground
#    37        set white foreground
#    38        256/24-bit foreground color follows, shoehorned into 16
#              basic colors (before Linux 3.16: set underscore on, set
#              default foreground color)
#    39        set default foreground color (before Linux 3.16: set
#              underscore off, set default foreground color)
#    40        set black background
#    41        set red background
#    42        set green background
#    43        set brown background
#    44        set blue background
#    45        set magenta background
#    46        set cyan background
#    47        set white background
#    48        256/24-bit background color follows, shoehorned into 8
#              basic colors
#    49        set default background color
#    90..97    set foreground to bright versions of 30..37
#    100..107  set background to bright versions of 40..47
#
#    Commands 38 and 48 require further arguments:
#      ;5;x       256 color: values 0..15 are IBGR (black, red, green,
#                 ... white), 16..231 a 6x6x6 color cube, 232..255 a
#                 grayscale ramp
#      ;2;r;g;b   24-bit color, r/g/b components are in the range 0..255
#
# For historical reasons, all "brown"s above are replaced with "yellow"
# by m.css.
#
# AnsiLexer supports commands 0, 1, 7, 22, 27, 30–39, 40–49, 90–97, and 100–107
# (ranges inclusive). All other commands will be ignored completely.
#
# Foreground colors named Bright* are not affected by the "bright" SGR
# setting—they will always appear bright, even after command 22 resets the
# normal intensity. Likewise, they will not affect the "bright" setting—a
# directive to use Red after BrightGreen will result in Red being displayed,
# not BrightRed.
#
# Palette or RGB Foreground colors from command 38 likewise do not interact
# with the "bright" setting, nor do any background colors.
class AnsiLexer(RegexLexer):
    name = 'Ansi escape lexer'

    _sgrs_and_text = '(?P<commands>(\x1b\\[(\\d*;)*\\d*m)+)(?P<text>[^\x1b]*)'
    _sgr_split = re.compile('m\x1b\\[|;|m|\x1b\\[')

    _named_colors = [
        'Black',
        'Red',
        'Green',
        'Yellow',
        'Blue',
        'Magenta',
        'Cyan',
        'White',
    ]
    _palette_start_colors = [
        '000000',
        '800000',
        '008000',
        '808000',
        '000080',
        '800080',
        '008080',
        'c0c0c0',
        '808080',
        'ff0000',
        '00ff00',
        'ffff00',
        '0000ff',
        'ff00ff',
        '00ffff',
        'ffffff',
    ]
    _palette_cube_steps = ['00', '5f', '87', 'af', 'd7', 'ff']

    def __init__(self, **options):
        RegexLexer.__init__(self, **options)

        self._bright = False
        self._invert = False
        self._foreground = ''
        self._background = ''

    def _callback(self, match):
        commands = match.group('commands')
        text = match.group('text')

        # split the commands strings into their constituent parameter codes
        parameters = self._sgr_split.split(commands)[1:-1]
        parameters = [int(p) if p else 0 for p in parameters]

        # loop over the parameters, consuming them to create commands, some
        # of which will have arguments
        while parameters:
            command = parameters.pop(0)
            if command == 0:
                self._bright = False
                self._invert = False
                self._foreground = ''
                self._background = ''
            elif command == 1:
                self._bright = True
            elif command == 7:
                self._invert = True
            elif command == 22:
                self._bright = False
            elif command == 27:
                self._invert = False
            elif command >= 30 and command <= 37:
                self._foreground = self._named_colors[command - 30]
            elif command == 38:
                mode = parameters.pop(0)
                if mode == 2:
                    rgb = self._read_rgb(parameters)
                else:
                    rgb = self._read_palette_color(parameters)
                self._foreground = rgb
            elif command == 39:
                self._foreground = ''
            elif command >= 40 and command <= 47:
                self._background = self._named_colors[command - 40]
            elif command == 48:
                mode = parameters.pop(0)
                if mode == 2:
                    rgb = self._read_rgb(parameters)
                else:
                    rgb = self._read_palette_color(parameters)
                self._background = rgb
            elif command == 49:
                self._background = ''
            elif command >= 90 and command <= 97:
                self._foreground = ('Bright' +
                                    self._named_colors[command - 90])
            elif command >= 100 and command <= 107:
                self._background = ('Bright' +
                                    self._named_colors[command - 100])

        # For "inverse video", foreground and background colors are swapped
        if self._invert:
            foreground, background = self._background, self._foreground
        else:
            foreground, background = self._foreground, self._background

        # If the bright bit is set for named colors (foreground or background
        # in case of reverse video), use their bright variant
        if self._bright and (not foreground or (foreground in self._named_colors)):
            token = 'Bright'
        else:
            token = ''

        # If the foreground color (or background in case of inverse video) is
        # set, use it. Otherwise pick a default if it's needed for the bright
        # variant, for inverse video or both.
        if foreground:
            token += foreground
        else:
            if self._invert:
                token += 'Inverted'
            if token:
                token += 'Default'

        # If the background color is set (or foreground in case of inverse
        # video), use it as well. This gets subsequently separated to either a
        # secondary CSS class or a separate background-color style item.
        if background:
            token += '-On' + background

        if token:
            token = 'Generic.Ansi' + token
            yield (match.start(), string_to_tokentype(token), text)
        else:
            yield (match.start(), Text, text)

    def _read_rgb(self, parameters):
        r = parameters.pop(0)
        g = parameters.pop(0)
        b = parameters.pop(0)
        return self._to_hex(r, g, b)

    def _read_palette_color(self, parameters):
        # the palette runs from 0–255 inclusive, consisting of
        #  - 16 specific colors intended to give a good range
        #  - 216 colors laid out on a color cube, with axes for
        #    each of red, green, and blue
        #  - 24 shades of grey, from grey3 to grey93
        offset = parameters.pop(0)
        if offset < 16:
            return self._palette_start_colors[offset]
        elif offset < 232:
            offset = offset - 16
            offset, b = divmod(offset, 6)
            r, g = divmod(offset, 6)
            r = self._palette_cube_steps[r]
            g = self._palette_cube_steps[g]
            b = self._palette_cube_steps[b]
            return r + g + b
        else:
            shade = 8 + 10 * (offset - 232)
            return self._to_hex(shade, shade, shade)

    def _to_hex(self, r, g, b):
        return '{:02x}{:02x}{:02x}'.format(r, g, b)

    tokens = {
        'root': [
            ('[^\x1b]+', Text),
            (_sgrs_and_text, _callback),
        ]
    }

class HtmlAnsiFormatter(HtmlFormatter):
    _ansi_color_re = re.compile(
        '(?P<Prefix>class="g g-Ansi)'
        '((?P<RGBForeground>[0-9a-f]{6})|(?P<NamedForeground>[A-Za-z]+))?'
        '(-On((?P<RGBBackground>[0-9a-f]{6})|(?P<NamedBackground>[A-Za-z]+)))?'
        '(?P<Suffix>")'
    )

    # Pygments 2.12+ doesn't have the outfile parameter anymore
    def wrap(self, source, outfile=None):
        return self._wrap_code(source)

    def _wrap_code(self, source):
        for i, t in source:
            if i == 1: # it's a line of formatted code
                t = self._ansi_color_re.sub(self._replace_ansi_class, t)
            yield i, t

    def _replace_ansi_class(self, match):
        html_classes = ['g']
        html_styles = []

        named_foreground = match.group('NamedForeground')
        if named_foreground:
            html_classes.append("g-Ansi" + named_foreground)
        else:
            rgb_foreground = match.group('RGBForeground')
            if rgb_foreground:
                html_styles.append('color: #' + rgb_foreground)

        named_background = match.group('NamedBackground')
        if named_background:
            html_classes.append("g-AnsiBackground" + named_background)
        else:
            rgb_background = match.group('RGBBackground')
            if rgb_background:
                html_styles.append('background-color: #' + rgb_background)

        result = ''
        if len(html_classes) > 1: # we don't want to emit just g
            result = 'class="' + ' '.join(html_classes) + '"'

        if html_styles:
            if result:
                result += ' '
            result += 'style="' + '; '.join(html_styles) + '"'

        return result



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re
from docutils import nodes, utils
from docutils.parsers import rst
from docutils.parsers.rst.roles import set_classes

# to avoid dependencies, link_regexp and parse_link() is common for m.abbr,
# m.gh, m.gl, m.link and m.vk

link_regexp = re.compile(r'(?P<title>.*) <(?P<link>.+)>')

def parse_link(text):
    link = utils.unescape(text)
    m = link_regexp.match(link)
    if m: return m.group('title', 'link')
    return None, link

def gh_internal(account, ref, title, link):
    base_url = "https://github.com/{}/{}/{}/{}"
    if '#' in ref:
        project, _, issue = ref.partition('#')
        url = base_url.format(account, project, "issues", issue)
        if not title: title = link
    elif '@' in ref:
        project, _, commit = ref.partition('@')
        url = base_url.format(account, project, "commit", commit)
        if not title: title = account + "/" + project + "@" + commit[0:7]
    elif '$' in ref:
        project, _, branch = ref.partition('$')
        url = base_url.format(account, project, "tree", branch)
        if not title: title = url
    elif '^' in ref:
        project, _, branch = ref.partition('^')
        url = base_url.format(account, project, "releases/tag", branch)
        if not title: title = url
    else:
        url = "https://github.com/{}/{}".format(account, ref)
        if not title:
            # if simple profile link, no need to expand to full URL
            title = link if not '/' in ref else url

    return title, url

def gh(name, rawtext, text, lineno, inliner, options={}, content=[]):
    title, link = parse_link(text)
    account, _, ref = link.partition('/')
    if not ref:
        url = "https://github.com/{}".format(account)
        if not title: title = "@{}".format(account)
    else:
        title, url = gh_internal(account, ref, title, link)

    set_classes(options)
    node = nodes.reference(rawtext, title, refuri=url, **options)
    return [node], []

def register_mcss(**kwargs):
    rst.roles.register_local_role('gh', gh)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

register = register_mcss # for Pelican
